"""
Boros HYPE snapshot helpers.

Kept as a mixin so the main strategy file stays readable without changing behavior.
"""

from __future__ import annotations

import time
from typing import Any

import aiohttp
from loguru import logger

from wayfinder_paths.core.utils.web3 import web3_from_chain_id

from .constants import (
    BOROS_HYPE_TOKEN_ID,
    ETH_ARB,
    HYPE_NATIVE,
    HYPE_OFT_ADDRESS,
    HYPEREVM_CHAIN_ID,
    KHYPE_API_URL,
    KHYPE_LST,
    KHYPE_STAKING_ACCOUNTANT,
    KHYPE_STAKING_ACCOUNTANT_ABI,
    LHYPE_ACCOUNTANT,
    LHYPE_ACCOUNTANT_ABI,
    LHYPE_API_URL,
    LOOPED_HYPE,
    MIN_HYPE_GAS,
    USDC_ARB,
    USDT_ARB,
    WHYPE,
    WHYPE_ADDRESS,
)
from .types import Inventory


async def fetch_lhype_apy() -> float | None:
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(LHYPE_API_URL, timeout=10) as resp:
                if resp.status == 200:
                    data = await resp.json()
                    if data.get("success") and data.get("result"):
                        reward_rate = data["result"].get("reward_rate")
                        if reward_rate is not None:
                            return float(reward_rate) / 100.0
    except Exception as e:
        logger.warning(f"Failed to fetch lHYPE APY: {e}")
    return None


async def fetch_khype_apy() -> float | None:
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(KHYPE_API_URL, timeout=10) as resp:
                if resp.status == 200:
                    data = await resp.json()
                    apy_14d = data.get("apy_14d")
                    if apy_14d is not None:
                        return float(apy_14d)
    except Exception as e:
        logger.warning(f"Failed to fetch kHYPE APY: {e}")
    return None


class BorosHypeSnapshotMixin:
    async def observe(self) -> Inventory:
        self._planner_runtime.reset_virtual_ledger()

        ok_addr, user_address = self._require_strategy_wallet_address()
        if not ok_addr:
            user_address = None

        # Default values
        hype_price_usd = 0.0
        hl_perp_margin = 0.0
        hl_spot_usdc = 0.0
        hl_spot_hype = 0.0
        hl_short_size_hype = 0.0
        hl_unrealized_pnl = 0.0
        hl_withdrawable_usd = 0.0
        mid_prices: dict[str, float] = {}
        perp_position: dict[str, Any] | None = None

        if self.hyperliquid_adapter and user_address:
            try:
                success, prices = await self.hyperliquid_adapter.get_all_mid_prices()
                if success and isinstance(prices, dict):
                    mid_prices = prices
                    hype_raw = prices.get("HYPE")
                    try:
                        hype_price_usd = (
                            float(hype_raw) if hype_raw is not None else 0.0
                        )
                    except (TypeError, ValueError):
                        hype_price_usd = 0.0

                success, user_state = await self.hyperliquid_adapter.get_user_state(
                    user_address
                )
                if success and isinstance(user_state, dict):
                    hl_perp_margin = self.hyperliquid_adapter.get_perp_margin_amount(
                        user_state
                    )
                    hl_withdrawable_usd = float(
                        user_state.get("withdrawable", 0)
                        or user_state.get("marginSummary", {}).get("totalRawUsd", 0)
                    )

                    positions = user_state.get("assetPositions", [])
                    for pos in positions:
                        pos_info = pos.get("position", {})
                        if pos_info.get("coin") == "HYPE":
                            perp_position = pos_info
                            szi = float(pos_info.get("szi", 0))
                            # Negative szi = short position
                            if szi < 0:
                                hl_short_size_hype = abs(szi)
                            hl_unrealized_pnl = float(pos_info.get("unrealizedPnl", 0))
                            break

                (
                    success,
                    spot_state,
                ) = await self.hyperliquid_adapter.get_spot_user_state(user_address)
                if success and isinstance(spot_state, dict):
                    balances = spot_state.get("balances", [])
                    for bal in balances:
                        token = bal.get("coin") or bal.get("token")
                        hold = float(bal.get("hold", 0))
                        total = float(bal.get("total", 0))
                        available = total - hold
                        if token == "USDC":
                            hl_spot_usdc = available
                        elif token == "HYPE":
                            hl_spot_hype = available

            except Exception as e:
                logger.warning(f"Failed to get Hyperliquid state: {e}")

        boros_idle_collateral_cross = 0.0
        boros_idle_collateral_isolated = 0.0
        boros_collateral_hype = 0.0
        boros_collateral_usd = 0.0
        boros_pending_withdrawal_hype = 0.0
        boros_pending_withdrawal_usd = 0.0
        boros_position_size = 0.0
        boros_position_value = 0.0
        boros_position_market_ids: set[int] = set()

        if self.boros_adapter:
            try:
                token_id = (
                    self._planner_runtime.current_boros_token_id or BOROS_HYPE_TOKEN_ID
                )
                success, balances = await self.boros_adapter.get_account_balances(
                    token_id=int(token_id)
                )
                if success and isinstance(balances, dict):
                    # Balances are returned in Boros cash units; for the HYPE-collateralized
                    # market these correspond to HYPE (18 decimals).
                    boros_collateral_hype = float(balances.get("total", 0))
                    boros_idle_collateral_cross = float(balances.get("cross", 0))
                    boros_idle_collateral_isolated = float(balances.get("isolated", 0))

                (
                    ok_pending,
                    pending_hype,
                ) = await self.boros_adapter.get_pending_withdrawal_amount(
                    token_id=int(token_id), token_decimals=18
                )
                if ok_pending:
                    boros_pending_withdrawal_hype = float(pending_hype)

                success, positions = await self.boros_adapter.get_active_positions()
                if success and isinstance(positions, list):
                    for pos in positions:
                        size = float(pos.get("size") or pos.get("notional", 0))
                        boros_position_size += abs(size)
                        boros_position_value += abs(size)
                        mid = pos.get("marketId") or pos.get("market_id")
                        try:
                            mid_int = int(mid) if mid is not None else None
                        except (TypeError, ValueError):
                            mid_int = None
                        if mid_int and mid_int > 0:
                            boros_position_market_ids.add(mid_int)

            except Exception as e:
                logger.warning(f"Failed to get Boros state: {e}")

        hype_hyperevm_balance = 0.0
        whype_balance = 0.0
        khype_balance = 0.0
        looped_hype_balance = 0.0
        usdc_arb_idle = 0.0
        usdt_arb_idle = 0.0
        eth_arb_balance = 0.0
        hype_oft_arb_balance = 0.0

        if self.balance_adapter:
            try:
                assets = [
                    {"token_id": HYPE_NATIVE},  # 0: HyperEVM native HYPE
                    {"token_id": WHYPE},  # 1: Wrapped HYPE
                    {"token_id": KHYPE_LST},  # 2: kHYPE
                    {"token_id": LOOPED_HYPE},  # 3: lHYPE
                    {"token_id": USDC_ARB},  # 4: Arbitrum USDC
                    {"token_id": USDT_ARB},  # 5: Arbitrum USDT
                    {"token_id": ETH_ARB},  # 6: Arbitrum ETH
                    {
                        "token_address": HYPE_OFT_ADDRESS,
                        "chain_id": 42161,
                    },  # 7: Arbitrum OFT HYPE
                ]
                ok, results = await self.balance_adapter.get_wallet_balances_multicall(
                    assets=assets
                )
                if ok and isinstance(results, list):
                    if results[0].get("success"):
                        hype_hyperevm_balance = results[0].get("balance_decimal") or 0.0
                    if results[1].get("success"):
                        whype_balance = results[1].get("balance_decimal") or 0.0
                    if results[2].get("success"):
                        khype_balance = results[2].get("balance_decimal") or 0.0
                    if results[3].get("success"):
                        looped_hype_balance = results[3].get("balance_decimal") or 0.0
                    if results[4].get("success"):
                        usdc_arb_idle = results[4].get("balance_decimal") or 0.0
                    if results[5].get("success"):
                        usdt_arb_idle = results[5].get("balance_decimal") or 0.0
                    if results[6].get("success"):
                        eth_arb_balance = results[6].get("balance_decimal") or 0.0
                    if results[7].get("success"):
                        hype_oft_arb_balance = results[7].get("balance_decimal") or 0.0

            except Exception as e:
                logger.warning(f"Failed to get wallet balances via multicall: {e}")

        # If we recently initiated a HyperEVM -> Arbitrum OFT bridge, the HYPE is
        # temporarily "in flight" (deducted from HyperEVM, not yet minted on Arb).
        # Track it in runtime to prevent hedge thrash + repeated funding.
        in_flight_hype = float(self._planner_runtime.in_flight_boros_oft_hype or 0.0)
        if in_flight_hype > 0:
            balance_before = float(
                self._planner_runtime.in_flight_boros_oft_hype_balance_before or 0.0
            )
            # Clear in-flight once Arb balance has increased by ~the bridged amount.
            if hype_oft_arb_balance >= balance_before + (in_flight_hype * 0.95):
                logger.info(
                    "Detected OFT HYPE arrival on Arbitrum; clearing in-flight bridge tracking"
                )
                self._planner_runtime.in_flight_boros_oft_hype = 0.0
                self._planner_runtime.in_flight_boros_oft_hype_balance_before = 0.0
                self._planner_runtime.in_flight_boros_oft_hype_started_at = None
                in_flight_hype = 0.0

        khype_to_hype_ratio = await self._get_khype_to_hype_ratio()
        looped_hype_to_hype_ratio = await self._get_looped_hype_to_hype_ratio()

        hl_spot_hype_value_usd = hl_spot_hype * hype_price_usd
        hype_hyperevm_value_usd = hype_hyperevm_balance * hype_price_usd
        whype_value_usd = whype_balance * hype_price_usd  # WHYPE is 1:1 with HYPE
        khype_value_usd = khype_balance * khype_to_hype_ratio * hype_price_usd
        looped_hype_value_usd = (
            looped_hype_balance * looped_hype_to_hype_ratio * hype_price_usd
        )
        hype_oft_arb_value_usd = hype_oft_arb_balance * hype_price_usd
        in_flight_hype_value_usd = in_flight_hype * hype_price_usd

        boros_collateral_usd = boros_collateral_hype * hype_price_usd
        boros_pending_withdrawal_usd = boros_pending_withdrawal_hype * hype_price_usd
        boros_committed_collateral_usd = (
            boros_collateral_usd + hype_oft_arb_value_usd + in_flight_hype_value_usd
        )

        # HyperEVM spot value (LSTs + native HYPE/WHYPE). Boros collateral and
        # Arbitrum OFT HYPE are tracked separately.
        spot_value_usd = (
            hype_hyperevm_value_usd
            + whype_value_usd
            + khype_value_usd
            + looped_hype_value_usd
            + hl_spot_hype_value_usd
        )

        # Gas reserve shouldn't be hedged; WHYPE counts as 1:1 HYPE exposure
        hedgeable_hyperevm_hype = max(0.0, hype_hyperevm_balance - MIN_HYPE_GAS)
        total_hype_exposure = (
            hedgeable_hyperevm_hype
            + whype_balance  # WHYPE is 1:1 with HYPE
            + (khype_balance * khype_to_hype_ratio)
            + (looped_hype_balance * looped_hype_to_hype_ratio)
            + hl_spot_hype
            + hype_oft_arb_balance
            + in_flight_hype
            + boros_collateral_hype
            + boros_pending_withdrawal_hype
        )

        hl_short_value_usd = hl_short_size_hype * hype_price_usd

        total_value = (
            spot_value_usd
            + hl_perp_margin
            + hl_spot_usdc
            + boros_collateral_usd
            + boros_pending_withdrawal_usd
            + usdc_arb_idle
            + usdt_arb_idle
            + hype_oft_arb_value_usd
            + in_flight_hype_value_usd
        )

        inv = Inventory(
            hype_hyperevm_balance=hype_hyperevm_balance,
            hype_hyperevm_value_usd=hype_hyperevm_value_usd,
            whype_balance=whype_balance,
            whype_value_usd=whype_value_usd,
            khype_balance=khype_balance,
            khype_value_usd=khype_value_usd,
            looped_hype_balance=looped_hype_balance,
            looped_hype_value_usd=looped_hype_value_usd,
            usdc_arb_idle=usdc_arb_idle,
            usdt_arb_idle=usdt_arb_idle,
            eth_arb_balance=eth_arb_balance,
            hype_oft_arb_balance=hype_oft_arb_balance,
            hype_oft_arb_value_usd=hype_oft_arb_value_usd,
            hl_perp_margin=hl_perp_margin,
            hl_spot_usdc=hl_spot_usdc,
            hl_spot_hype=hl_spot_hype,
            hl_spot_hype_value_usd=hl_spot_hype_value_usd,
            hl_short_size_hype=hl_short_size_hype,
            hl_short_value_usd=hl_short_value_usd,
            hl_unrealized_pnl=hl_unrealized_pnl,
            hl_withdrawable_usd=hl_withdrawable_usd,
            boros_idle_collateral_isolated=boros_idle_collateral_isolated,
            boros_idle_collateral_cross=boros_idle_collateral_cross,
            boros_collateral_hype=boros_collateral_hype,
            boros_collateral_usd=boros_collateral_usd,
            boros_pending_withdrawal_hype=boros_pending_withdrawal_hype,
            boros_pending_withdrawal_usd=boros_pending_withdrawal_usd,
            boros_committed_collateral_usd=boros_committed_collateral_usd,
            boros_position_size=boros_position_size,
            boros_position_value=boros_position_value,
            khype_to_hype_ratio=khype_to_hype_ratio,
            looped_hype_to_hype_ratio=looped_hype_to_hype_ratio,
            hype_price_usd=hype_price_usd,
            spot_value_usd=spot_value_usd,
            total_hype_exposure=total_hype_exposure,
            total_value=total_value,
            boros_position_market_ids=sorted(boros_position_market_ids)
            if boros_position_market_ids
            else None,
        )

        self._opa_alloc = self._get_allocation_status(inv)

        self._opa_risk_progress = self._hyperliquid_liquidation_progress(
            perp_position, mid_prices
        )

        # Only set pending flag for actual Boros withdrawal (not for idle USDT on Arb)
        if inv.boros_pending_withdrawal_usd > 1.0:
            self._opa_pending_withdrawal = True

        # Check for HL liquidation only when hedge is gone but spot exposure exists
        has_no_short = abs(inv.hl_short_size_hype) < 0.01
        has_spot_exposure = inv.total_hype_exposure > 0.1

        if (
            has_no_short
            and has_spot_exposure
            and self.hyperliquid_adapter
            and user_address
        ):
            since_ms = int((time.time() - 43200) * 1000)  # last 12 hours
            try:
                (
                    ok,
                    liq_fills,
                ) = await self.hyperliquid_adapter.check_recent_liquidations(
                    user_address, since_ms
                )
                if ok and liq_fills:
                    inv.hl_liquidation_detected = True
                    inv.hl_liquidation_fills = liq_fills
                    logger.warning(
                        f"[LIQUIDATION] HL position was liquidated! "
                        f"Short={inv.hl_short_size_hype:.4f}, "
                        f"Spot exposure={inv.total_hype_exposure:.4f}"
                    )
                    for fill in liq_fills:
                        liq = fill.get("liquidation", {})
                        logger.warning(
                            f"[LIQUIDATION] coin={fill.get('coin')}, sz={fill.get('sz')}, "
                            f"method={liq.get('method')}, markPx={liq.get('markPx')}"
                        )
            except Exception as e:
                logger.warning(f"Failed to check for liquidations: {e}")

        if self.boros_adapter:
            try:
                success, quotes = await self.boros_adapter.quote_markets_for_underlying(
                    "HYPE"
                )
                if success:
                    self._opa_boros_quotes = quotes
                    logger.debug(f"Fetched {len(quotes)} Boros HYPE quotes")
            except Exception as e:
                logger.warning(f"Failed to get Boros quotes: {e}")
                self._opa_boros_quotes = []

        return inv

    def _hyperliquid_liquidation_progress(
        self,
        perp_pos: dict[str, Any] | None,
        mid_prices: dict[str, float] | None = None,
    ) -> float:
        # Returns fraction [0,1] of distance from entry to liquidation (0 = at entry, 1 = at liq)
        if not perp_pos:
            return 0.0

        liq_px = perp_pos.get("liquidationPx") or perp_pos.get("liqPx")
        entry_px = perp_pos.get("entryPx") or perp_pos.get("entryPrice")
        szi = perp_pos.get("szi") or perp_pos.get("size")
        coin = perp_pos.get("coin", "HYPE")

        mark_px = None
        if mid_prices:
            mark_px = mid_prices.get(coin)
        if mark_px is None:
            mark_px = perp_pos.get("px") or perp_pos.get("markPx")

        if not all([liq_px, entry_px, mark_px, szi]):
            return 0.0

        try:
            liq = float(liq_px)
            entry = float(entry_px)
            mark = float(mark_px)
            size = float(szi)

            if abs(liq - entry) < 0.0001:
                return 0.0

            # For SHORT positions (szi < 0):
            # Progress = (mark - entry) / (liq - entry)
            # When mark rises toward liq, progress → 1
            if size < 0:
                progress = (mark - entry) / (liq - entry)
            else:
                # For LONG positions (szi > 0):
                # Progress = (entry - mark) / (entry - liq)
                # When mark falls toward liq, progress → 1
                progress = (entry - mark) / (entry - liq)

            return max(0.0, min(1.0, progress))
        except (ValueError, ZeroDivisionError):
            return 0.0

    async def _get_khype_to_hype_ratio(self) -> float:
        # Query Kinetiq StakingAccountant kHYPEToHYPE(1e18) for HYPE per 1 kHYPE
        try:
            async with web3_from_chain_id(HYPEREVM_CHAIN_ID) as w3:
                # kHYPE has 18 decimals, so 1 kHYPE = 1e18
                one_khype = 10**18

                contract = w3.eth.contract(
                    address=w3.to_checksum_address(KHYPE_STAKING_ACCOUNTANT),
                    abi=KHYPE_STAKING_ACCOUNTANT_ABI,
                )
                hype_raw = await contract.functions.kHYPEToHYPE(one_khype).call()

                # HYPE also has 18 decimals
                return int(hype_raw) / (10**18)
        except Exception as e:
            logger.warning(f"Failed to get kHYPE exchange rate: {e}")
            return 1.0  # Default to 1:1

    async def _get_looped_hype_to_hype_ratio(self) -> float:
        # Query Looping Accountant getRateInQuote(WHYPE) for HYPE per 1 LHYPE
        try:
            async with web3_from_chain_id(HYPEREVM_CHAIN_ID) as w3:
                contract = w3.eth.contract(
                    address=w3.to_checksum_address(LHYPE_ACCOUNTANT),
                    abi=LHYPE_ACCOUNTANT_ABI,
                )
                rate_raw = await contract.functions.getRateInQuote(
                    w3.to_checksum_address(WHYPE_ADDRESS)
                ).call()

                # Rate is returned with 18 decimals (WHYPE decimals)
                return int(rate_raw) / (10**18)
        except Exception as e:
            logger.warning(f"Failed to get LHYPE exchange rate: {e}")
            return 1.0  # Default to 1:1



"""LayerZero OFT ABI subset used for bridging native HYPE from HyperEVM."""

HYPE_OFT_ABI = [
    {
        "inputs": [
            {
                "components": [
                    {"internalType": "uint32", "name": "dstEid", "type": "uint32"},
                    {"internalType": "bytes32", "name": "to", "type": "bytes32"},
                    {"internalType": "uint256", "name": "amountLD", "type": "uint256"},
                    {
                        "internalType": "uint256",
                        "name": "minAmountLD",
                        "type": "uint256",
                    },
                    {
                        "internalType": "bytes",
                        "name": "extraOptions",
                        "type": "bytes",
                    },
                    {"internalType": "bytes", "name": "composeMsg", "type": "bytes"},
                    {"internalType": "bytes", "name": "oftCmd", "type": "bytes"},
                ],
                "internalType": "struct SendParam",
                "name": "_sendParam",
                "type": "tuple",
            },
            {"internalType": "bool", "name": "_payInLzToken", "type": "bool"},
        ],
        "name": "quoteSend",
        "outputs": [
            {
                "components": [
                    {"internalType": "uint256", "name": "nativeFee", "type": "uint256"},
                    {
                        "internalType": "uint256",
                        "name": "lzTokenFee",
                        "type": "uint256",
                    },
                ],
                "internalType": "struct MessagingFee",
                "name": "",
                "type": "tuple",
            }
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {
                "components": [
                    {"internalType": "uint32", "name": "dstEid", "type": "uint32"},
                    {"internalType": "bytes32", "name": "to", "type": "bytes32"},
                    {"internalType": "uint256", "name": "amountLD", "type": "uint256"},
                    {
                        "internalType": "uint256",
                        "name": "minAmountLD",
                        "type": "uint256",
                    },
                    {
                        "internalType": "bytes",
                        "name": "extraOptions",
                        "type": "bytes",
                    },
                    {"internalType": "bytes", "name": "composeMsg", "type": "bytes"},
                    {"internalType": "bytes", "name": "oftCmd", "type": "bytes"},
                ],
                "internalType": "struct SendParam",
                "name": "_sendParam",
                "type": "tuple",
            },
            {
                "components": [
                    {"internalType": "uint256", "name": "nativeFee", "type": "uint256"},
                    {
                        "internalType": "uint256",
                        "name": "lzTokenFee",
                        "type": "uint256",
                    },
                ],
                "internalType": "struct MessagingFee",
                "name": "_fee",
                "type": "tuple",
            },
            {"internalType": "address", "name": "_refundAddress", "type": "address"},
        ],
        "name": "send",
        "outputs": [
            {
                "components": [
                    {"internalType": "bytes32", "name": "guid", "type": "bytes32"},
                    {"internalType": "uint64", "name": "nonce", "type": "uint64"},
                    {
                        "components": [
                            {
                                "internalType": "uint256",
                                "name": "nativeFee",
                                "type": "uint256",
                            },
                            {
                                "internalType": "uint256",
                                "name": "lzTokenFee",
                                "type": "uint256",
                            },
                        ],
                        "internalType": "struct MessagingFee",
                        "name": "fee",
                        "type": "tuple",
                    },
                ],
                "internalType": "struct MessagingReceipt",
                "name": "",
                "type": "tuple",
            },
            {
                "components": [
                    {
                        "internalType": "uint256",
                        "name": "amountSentLD",
                        "type": "uint256",
                    },
                    {
                        "internalType": "uint256",
                        "name": "amountReceivedLD",
                        "type": "uint256",
                    },
                ],
                "internalType": "struct OFTReceipt",
                "name": "",
                "type": "tuple",
            },
        ],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "sharedDecimals",
        "outputs": [{"internalType": "uint8", "name": "", "type": "uint8"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "decimalConversionRate",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
]



from unittest.mock import AsyncMock, patch

import pytest

from wayfinder_paths.adapters.balance_adapter.adapter import BalanceAdapter
from wayfinder_paths.core.utils.token_resolver import TokenResolver


@pytest.fixture(autouse=True)
def _clear_token_resolver_cache():
    TokenResolver._token_details_cache.clear()
    TokenResolver._gas_token_cache.clear()


class TestBalanceAdapter:
    @pytest.fixture
    def adapter(self):
        return BalanceAdapter(config={})

    def test_adapter_type(self, adapter):
        assert adapter.adapter_type == "BALANCE"

    @pytest.mark.asyncio
    async def test_get_balance_with_token_id(self, adapter):
        token_address = "0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913"

        with (
            patch(
                "wayfinder_paths.adapters.balance_adapter.adapter.TokenResolver.resolve_token",
                new_callable=AsyncMock,
                return_value=(8453, token_address),
            ) as mock_resolve,
            patch(
                "wayfinder_paths.adapters.balance_adapter.adapter.get_token_balance",
                new_callable=AsyncMock,
                return_value=1_000_000,
            ) as mock_get_balance,
        ):
            success, balance = await adapter.get_balance(
                token_id="usd-coin-base",
                wallet_address="0xWallet",
            )

            assert success is True
            assert balance == 1_000_000
            mock_resolve.assert_called_once_with("usd-coin-base", chain_id=None)
            mock_get_balance.assert_called_once_with(token_address, 8453, "0xWallet")

    @pytest.mark.asyncio
    async def test_get_balance_with_token_address(self, adapter):
        with patch(
            "wayfinder_paths.adapters.balance_adapter.adapter.get_token_balance",
            new_callable=AsyncMock,
            return_value=5_000_000,
        ) as mock_get_balance:
            success, balance = await adapter.get_balance(
                token_address="0xTokenAddress",
                wallet_address="0xWallet",
                chain_id=8453,
            )

            assert success is True
            assert balance == 5_000_000
            mock_get_balance.assert_called_once_with("0xTokenAddress", 8453, "0xWallet")

    @pytest.mark.asyncio
    async def test_get_balance_token_not_found(self, adapter):
        with patch(
            "wayfinder_paths.adapters.balance_adapter.adapter.TokenResolver.resolve_token",
            new_callable=AsyncMock,
            side_effect=ValueError("Cannot resolve token: invalid-token"),
        ):
            success, error = await adapter.get_balance(
                token_id="invalid-token",
                wallet_address="0xWallet",
            )

            assert success is False
            assert "Cannot resolve token" in str(error)

    @pytest.mark.asyncio
    async def test_get_balance_parses_address_token_id_locally(self, adapter):
        token_id = "base_0x1111111111111111111111111111111111111111"

        with (
            patch(
                "wayfinder_paths.core.utils.token_resolver.TOKEN_CLIENT.get_token_details",
                new=AsyncMock(),
            ) as mock_token_details,
            patch(
                "wayfinder_paths.adapters.balance_adapter.adapter.get_token_balance",
                new_callable=AsyncMock,
                return_value=123,
            ) as mock_get_balance,
        ):
            success, balance = await adapter.get_balance(
                token_id=token_id,
                wallet_address="0xWallet",
            )

            assert success is True
            assert balance == 123
            mock_token_details.assert_not_called()
            mock_get_balance.assert_called_once_with(
                "0x1111111111111111111111111111111111111111", 8453, "0xWallet"
            )

    @pytest.mark.asyncio
    async def test_get_balance_details_fetches_balance_and_decimals(self, adapter):
        token_address = "0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913"

        with (
            patch(
                "wayfinder_paths.adapters.balance_adapter.adapter.TokenResolver.resolve_token",
                new_callable=AsyncMock,
                return_value=(8453, token_address),
            ) as mock_resolve,
            patch(
                "wayfinder_paths.adapters.balance_adapter.adapter.get_token_balance_with_decimals",
                new_callable=AsyncMock,
                return_value=(1_000_000, 6),
            ) as mock_get_balance,
        ):
            ok, out = await adapter.get_balance_details(
                token_id="usd-coin-base",
                wallet_address="0xWallet",
            )

            assert ok is True
            assert isinstance(out, dict)
            assert out["balance_raw"] == 1_000_000
            assert out["decimals"] == 6
            assert out["balance_decimal"] == 1.0

            mock_resolve.assert_called_once_with("usd-coin-base", chain_id=None)
            mock_get_balance.assert_called_once_with(
                token_address,
                8453,
                "0xWallet",
                balance_block_identifier="pending",
                default_native_decimals=18,
            )

    @pytest.mark.asyncio
    async def test_get_balance_details_falls_back_to_onchain_decimals(self, adapter):
        token_id = "base_0x1111111111111111111111111111111111111111"

        with (
            patch(
                "wayfinder_paths.core.utils.token_resolver.TOKEN_CLIENT.get_token_details",
                new=AsyncMock(),
            ) as mock_token_details,
            patch(
                "wayfinder_paths.adapters.balance_adapter.adapter.get_token_balance_with_decimals",
                new_callable=AsyncMock,
                return_value=(1_000_000, 6),
            ) as mock_get_balance,
        ):
            ok, out = await adapter.get_balance_details(
                token_id=token_id,
                wallet_address="0xWallet",
            )

            assert ok is True
            assert isinstance(out, dict)
            assert out["balance_raw"] == 1_000_000
            assert out["decimals"] == 6
            assert out["balance_decimal"] == 1.0

            mock_token_details.assert_not_called()
            mock_get_balance.assert_called_once_with(
                "0x1111111111111111111111111111111111111111",
                8453,
                "0xWallet",
                balance_block_identifier="pending",
                default_native_decimals=18,
            )



from __future__ import annotations

import unicodedata

SYMBOL_TRANSLATION_TABLE = str.maketrans(
    {
        "₮": "T",
        "₿": "B",
        "Ξ": "X",
    }
)

STABLE_SYMBOL_KEYWORDS = {
    "USD",
    "USDC",
    "USDT",
    "USDP",
    "USDD",
    "USDS",
    "DAI",
    "USKB",
    "USDE",
    "USDH",
    "USDL",
    "USDR",
    "USDX",
    "SUSD",
    "LUSD",
    "GUSD",
    "TUSD",
    "USR",
    "USDHL",
}


def normalize_symbol(symbol: str | None) -> str:
    """Normalize a token symbol for stable comparisons / keys.

    - Unicode NFKD normalization
    - Common crypto symbol translation (₮/₿/Ξ)
    - ASCII-only
    - Keep only alphanumerics
    - Lowercase
    """
    if symbol is None:
        return ""
    normalized = unicodedata.normalize("NFKD", str(symbol)).translate(
        SYMBOL_TRANSLATION_TABLE
    )
    ascii_only = normalized.encode("ascii", "ignore").decode("ascii")
    filtered = "".join(ch for ch in ascii_only if ch.isalnum())
    if filtered:
        return filtered.lower()
    return str(symbol).lower()


def is_stable_symbol(symbol: str | None) -> bool:
    if not symbol:
        return False
    upper = str(symbol).upper()
    return any(keyword in upper for keyword in STABLE_SYMBOL_KEYWORDS)



import time
from pathlib import Path
from unittest.mock import AsyncMock

import pytest

from wayfinder_paths.core.constants.projectx import THBILL_USDC_METADATA
from wayfinder_paths.core.utils.uniswap_v3_math import price_to_sqrt_price_x96
from wayfinder_paths.strategies.projectx_thbill_usdc_strategy.strategy import (
    ProjectXThbillUsdcStrategy,
)
from wayfinder_paths.tests.test_utils import (
    assert_status_dict,
    assert_status_tuple,
    load_strategy_examples,
)


@pytest.fixture
def strategy():
    mock_config = {
        "main_wallet": {"address": "0x1234567890123456789012345678901234567890"},
        "strategy_wallet": {"address": "0xabcdefabcdefabcdefabcdefabcdefabcdefabcd"},
    }
    s = ProjectXThbillUsdcStrategy(config=mock_config)

    # Avoid real network calls
    s.token_adapter.get_token = AsyncMock(return_value=(True, {"decimals": 6}))
    s.token_adapter.get_token_price = AsyncMock(
        return_value=(True, {"current_price": 1.0})
    )

    def get_balance_side_effect(*, wallet_address, token_id=None, **kwargs):
        if token_id == "usd-coin-hyperevm":
            return (True, 10_000_000)  # 10 USDC @ 6dp
        if token_id == "hype-hyperevm":
            return (True, 2_000_000_000_000_000_000)  # 2 HYPE @ 18dp
        return (True, 0)

    s.balance_adapter.get_balance = AsyncMock(side_effect=get_balance_side_effect)
    s.balance_adapter.move_from_main_wallet_to_strategy_wallet = AsyncMock(
        return_value=(True, "0xtxhash_transfer")
    )
    s.balance_adapter.move_from_strategy_wallet_to_main_wallet = AsyncMock(
        return_value=(True, "0xtxhash_transfer")
    )

    s.ledger_adapter.get_strategy_net_deposit = AsyncMock(return_value=(True, 0.0))

    s.projectx.pool_overview = AsyncMock(
        return_value=(
            True,
            {
                "sqrt_price_x96": 0,
                "tick": 0,
                "tick_spacing": 10,
                "fee": 100,
                "token0": {
                    "address": "0x0",
                    "decimals": 6,
                    "symbol": "USDC",
                    "token_id": "usd-coin-hyperevm",
                },
                "token1": {
                    "address": "0x1",
                    "decimals": 18,
                    "symbol": "THBILL",
                    "token_id": "theo-short-duration-us-treasury-fund-hyperevm",
                },
            },
        )
    )
    s.projectx.list_positions = AsyncMock(return_value=(True, []))
    s.projectx.mint_from_balances = AsyncMock(
        return_value=(
            True,
            {
                "token_id": 123,
                "tx_hash": "0xtxhash_mint",
                "spent": {"token0_spent": 0, "token1_spent": 0},
            },
        )
    )
    s.projectx.current_balances = AsyncMock(return_value=(True, {"0x0": 0, "0x1": 0}))
    s.projectx.fetch_swaps = AsyncMock(return_value=(True, []))
    s.projectx.fetch_prjx_points = AsyncMock(return_value=(True, {"points": 0}))
    s.projectx.live_fee_snapshot = AsyncMock(return_value=(True, {"usd": 0.0}))

    return s


def test_descriptor_includes_point_rewards():
    rewards = ProjectXThbillUsdcStrategy.INFO.available_rewards or {}
    expected_program = THBILL_USDC_METADATA.get("points_program")
    if expected_program:
        assert rewards.get("point_rewards")
        assert rewards["point_rewards"][0]["program"] == expected_program
    else:
        assert rewards.get("point_rewards") is None


@pytest.mark.asyncio
@pytest.mark.smoke
async def test_smoke(strategy):
    examples = load_strategy_examples(Path(__file__))
    deposit_args = examples.get("smoke", {}).get("deposit", {})

    st = await strategy.status()
    assert_status_dict(st)

    ok, msg = await strategy.deposit(**deposit_args)
    assert_status_tuple((ok, msg))

    ok, msg = await strategy.update()
    assert_status_tuple((ok, msg))

    ok, msg = await strategy.exit()
    assert_status_tuple((ok, msg))


@pytest.mark.asyncio
async def test_quote_estimates_fee_apy_from_swap_volume(strategy):
    sqrt_price_x96 = price_to_sqrt_price_x96(1.0, 6, 18)
    strategy.projectx.pool_overview = AsyncMock(
        return_value=(
            True,
            {
                "sqrt_price_x96": sqrt_price_x96,
                "tick": 0,
                "tick_spacing": 10,
                "fee": 3000,  # 0.30%
                "liquidity": 0,  # simplify share -> 1.0
                "token0": {
                    "address": "0x0",
                    "decimals": 6,
                    "symbol": "USDC",
                    "token_id": "usd-coin-hyperevm",
                },
                "token1": {
                    "address": "0x1",
                    "decimals": 18,
                    "symbol": "THBILL",
                    "token_id": "theo-short-duration-us-treasury-fund-hyperevm",
                },
            },
        )
    )

    strategy.projectx.fetch_swaps = AsyncMock(
        return_value=(
            True,
            [
                {"timestamp": int(time.time()), "tick": 0, "amount_usd": 10_000.0},
                {"timestamp": int(time.time()), "tick": 10, "amount_usd": 5_000.0},
                {"timestamp": int(time.time()), "tick": 100, "amount_usd": 20_000.0},
            ],
        )
    )

    q = await strategy.quote(deposit_amount=1000.0)
    assert q["expected_apy"] == pytest.approx(16.425, rel=1e-6)
    assert q["components"]["volume_usd_in_range"] == pytest.approx(15_000.0, rel=1e-9)



import json
import os
from pathlib import Path
from typing import Any

_CONFIG_ENV_KEYS = ("WAYFINDER_CONFIG_PATH", "WAYFINDER_CONFIG")
_DEFAULT_CONFIG_FILENAME = "config.json"
_WALLET_MNEMONIC_KEY = "wallet_mnemonic"


def _find_project_root(start: Path) -> Path | None:
    cur = start.resolve()
    for parent in [cur, *cur.parents]:
        if (parent / "pyproject.toml").exists():
            return parent
    return None


def _project_root() -> Path | None:
    return _find_project_root(Path.cwd()) or _find_project_root(Path(__file__).parent)


def resolve_config_path(path: str | Path | None = None) -> Path:
    if path is not None:
        return Path(path).expanduser()

    env_path = next(
        (os.getenv(k, "").strip() for k in _CONFIG_ENV_KEYS if os.getenv(k)), ""
    )
    if env_path:
        p = Path(env_path).expanduser()
        if p.is_absolute():
            return p
        root = _project_root()
        return (root / p) if root else p

    root = _project_root()
    return (root / _DEFAULT_CONFIG_FILENAME) if root else Path(_DEFAULT_CONFIG_FILENAME)


def load_config_json(
    path: str | Path | None = None, *, require_exists: bool = False
) -> dict[str, Any]:
    cfg_path = resolve_config_path(path)
    if not cfg_path.exists():
        if require_exists:
            raise FileNotFoundError(f"Config file not found: {cfg_path}")
        return {}
    try:
        return json.loads(cfg_path.read_text())
    except Exception:
        return {}


def write_config_json(path: str | Path | None, config: dict[str, Any]) -> Path:
    cfg_path = resolve_config_path(path)
    cfg_path.parent.mkdir(parents=True, exist_ok=True)
    cfg_path.write_text(json.dumps(config, indent=2) + "\n")
    return cfg_path


CONFIG: dict[str, Any] = load_config_json()


def set_config(config: dict[str, Any]) -> None:
    """Replace the global CONFIG dict in-place.

    This allows code that imported CONFIG at module import time to see updates.
    """
    CONFIG.clear()
    CONFIG.update(config)


def load_config(
    path: str | Path | None = None, *, require_exists: bool = False
) -> None:
    """Load config from disk into the global CONFIG dict."""
    set_config(load_config_json(path, require_exists=require_exists))


def set_rpc_urls(rpc_urls):
    if "strategy" not in CONFIG:
        CONFIG["strategy"] = {}
    if "rpc_urls" not in CONFIG["strategy"]:
        CONFIG["strategy"]["rpc_urls"] = {}
    CONFIG["strategy"]["rpc_urls"] = rpc_urls


def get_rpc_urls() -> dict[str, Any]:
    return CONFIG.get("strategy", {}).get("rpc_urls", {})


def get_api_base_url() -> str:
    system = CONFIG.get("system", {}) if isinstance(CONFIG, dict) else {}
    api_url = system.get("api_base_url")
    if api_url and isinstance(api_url, str):
        return api_url.strip()
    return "https://wayfinder.ai/api"


def get_api_key() -> str | None:
    system = CONFIG.get("system", {}) if isinstance(CONFIG, dict) else {}
    api_key = system.get("api_key")
    if api_key and isinstance(api_key, str):
        return api_key.strip()
    return os.environ.get("WAYFINDER_API_KEY")


def load_wallet_mnemonic(path: str | Path | None = None) -> str | None:
    config = CONFIG if path is None else load_config_json(path)
    value = config.get(_WALLET_MNEMONIC_KEY)
    if isinstance(value, str) and value.strip():
        return value.strip()
    return None


def write_wallet_mnemonic(mnemonic: str, path: str | Path | None = None) -> Path:
    cfg_path = resolve_config_path(path)
    config = load_config_json(cfg_path)
    config[_WALLET_MNEMONIC_KEY] = mnemonic
    write_config_json(cfg_path, config)

    default_path = resolve_config_path()
    if cfg_path.resolve() == default_path.resolve():
        CONFIG[_WALLET_MNEMONIC_KEY] = mnemonic

    return cfg_path


_GORLAMI_PATH = "/blockchain/gorlami"


def get_gorlami_base_url() -> str:
    return f"{get_api_base_url().rstrip('/')}{_GORLAMI_PATH}"



"""Verify all strategies conform to expected interface."""

import inspect
from importlib import import_module
from pathlib import Path

import pytest

from wayfinder_paths.core.strategies import Strategy


def get_all_strategy_classes():
    """Discover all Strategy subclasses in wayfinder_paths/strategies/."""
    strategies_dir = Path(__file__).parent.parent / "strategies"
    strategy_classes = []

    for strategy_dir in strategies_dir.iterdir():
        if not strategy_dir.is_dir() or strategy_dir.name.startswith("_"):
            continue
        strategy_file = strategy_dir / "strategy.py"
        if not strategy_file.exists():
            continue

        module_path = f"wayfinder_paths.strategies.{strategy_dir.name}.strategy"
        try:
            module = import_module(module_path)
            for _name, obj in inspect.getmembers(module, inspect.isclass):
                if issubclass(obj, Strategy) and obj is not Strategy:
                    strategy_classes.append((strategy_dir.name, obj))
        except ImportError:
            continue

    return strategy_classes


@pytest.mark.parametrize("strategy_name,strategy_class", get_all_strategy_classes())
def test_deposit_accepts_run_strategy_kwargs(strategy_name, strategy_class):
    """
    Verify deposit() accepts main_token_amount and gas_token_amount kwargs.

    Both run_strategy.py and the MCP tool runner pass these kwargs.
    """
    sig = inspect.signature(strategy_class.deposit)
    params = sig.parameters

    # Must accept **kwargs OR explicitly include the canonical params.
    has_var_keyword = any(
        p.kind == inspect.Parameter.VAR_KEYWORD for p in params.values()
    )
    if has_var_keyword:
        return

    assert "main_token_amount" in params, (
        f"{strategy_class.__name__}.deposit() must accept main_token_amount "
        "or **kwargs."
    )
    assert "gas_token_amount" in params, (
        f"{strategy_class.__name__}.deposit() must accept gas_token_amount or **kwargs."
    )


@pytest.mark.parametrize("strategy_name,strategy_class", get_all_strategy_classes())
def test_withdraw_accepts_run_strategy_kwargs(strategy_name, strategy_class):
    """
    Verify withdraw() accepts max_wait_s and poll_interval_s kwargs.

    run_strategy.py always passes these to withdraw(), so all strategies
    must accept **kwargs to avoid TypeError.
    """
    sig = inspect.signature(strategy_class.withdraw)
    params = sig.parameters

    # Must accept **kwargs (VAR_KEYWORD)
    has_var_keyword = any(
        p.kind == inspect.Parameter.VAR_KEYWORD for p in params.values()
    )

    assert has_var_keyword, (
        f"{strategy_class.__name__}.withdraw() must accept **kwargs. "
        f"run_strategy.py passes max_wait_s and poll_interval_s to all strategies."
    )


@pytest.mark.parametrize("strategy_name,strategy_class", get_all_strategy_classes())
def test_quote_accepts_deposit_amount(strategy_name, strategy_class):
    """
    Verify quote() accepts deposit_amount kwarg when implemented.

    run_strategy.py and the MCP tool runner call quote(deposit_amount=...).
    """
    quote_fn = getattr(strategy_class, "quote", None)
    if not callable(quote_fn):
        return

    sig = inspect.signature(quote_fn)
    params = sig.parameters

    has_var_keyword = any(
        p.kind == inspect.Parameter.VAR_KEYWORD for p in params.values()
    )
    if has_var_keyword:
        return

    assert "deposit_amount" in params, (
        f"{strategy_class.__name__}.quote() must accept deposit_amount or **kwargs."
    )


@pytest.mark.parametrize("strategy_name,strategy_class", get_all_strategy_classes())
def test_analyze_accepts_cli_kwargs(strategy_name, strategy_class):
    """
    Verify analyze() accepts deposit_usdc and verbose kwargs when implemented.

    run_strategy.py calls analyze(deposit_usdc=..., verbose=...).
    """
    analyze_fn = getattr(strategy_class, "analyze", None)
    if not callable(analyze_fn):
        return

    sig = inspect.signature(analyze_fn)
    params = sig.parameters

    has_var_keyword = any(
        p.kind == inspect.Parameter.VAR_KEYWORD for p in params.values()
    )
    if has_var_keyword:
        return

    assert "deposit_usdc" in params, (
        f"{strategy_class.__name__}.analyze() must accept deposit_usdc or **kwargs."
    )
    assert "verbose" in params, (
        f"{strategy_class.__name__}.analyze() must accept verbose or **kwargs."
    )


@pytest.mark.parametrize("strategy_name,strategy_class", get_all_strategy_classes())
def test_build_batch_snapshot_accepts_score_deposit_usdc(strategy_name, strategy_class):
    """
    Verify build_batch_snapshot() accepts score_deposit_usdc kwarg when implemented.

    The MCP tool runner calls build_batch_snapshot(score_deposit_usdc=...).
    """
    snapshot_fn = getattr(strategy_class, "build_batch_snapshot", None)
    if not callable(snapshot_fn):
        return

    sig = inspect.signature(snapshot_fn)
    params = sig.parameters

    has_var_keyword = any(
        p.kind == inspect.Parameter.VAR_KEYWORD for p in params.values()
    )
    if has_var_keyword:
        return

    assert "score_deposit_usdc" in params, (
        f"{strategy_class.__name__}.build_batch_snapshot() must accept "
        "score_deposit_usdc or **kwargs."
    )



from __future__ import annotations

from typing import Any, Literal

from eth_utils import to_checksum_address

from wayfinder_paths.adapters.ledger_adapter.adapter import LedgerAdapter
from wayfinder_paths.adapters.token_adapter.adapter import TokenAdapter
from wayfinder_paths.core.adapters.BaseAdapter import BaseAdapter
from wayfinder_paths.core.adapters.models import LEND, UNLEND
from wayfinder_paths.core.clients.HyperlendClient import (
    HYPERLEND_CLIENT,
    AssetsView,
    LendRateHistory,
    MarketEntry,
    StableMarketsHeadroomResponse,
)
from wayfinder_paths.core.constants.base import MAX_UINT256
from wayfinder_paths.core.constants.chains import CHAIN_ID_HYPEREVM
from wayfinder_paths.core.constants.contracts import (
    HYPEREVM_WHYPE,
    HYPERLEND_POOL,
    HYPERLEND_POOL_ADDRESSES_PROVIDER,
    HYPERLEND_UI_POOL_DATA_PROVIDER,
    HYPERLEND_WRAPPED_TOKEN_GATEWAY,
)
from wayfinder_paths.core.constants.hyperlend_abi import (
    POOL_ABI,
    UI_POOL_DATA_PROVIDER_ABI,
    UI_POOL_RESERVE_KEYS,
    WETH_ABI,
    WRAPPED_TOKEN_GATEWAY_ABI,
)
from wayfinder_paths.core.utils.interest import RAY, apr_to_apy, ray_to_apr
from wayfinder_paths.core.utils.symbols import is_stable_symbol, normalize_symbol
from wayfinder_paths.core.utils.tokens import ensure_allowance, get_token_balance
from wayfinder_paths.core.utils.transaction import encode_call, send_transaction
from wayfinder_paths.core.utils.web3 import web3_from_chain_id

VARIABLE_RATE_MODE = 2
REFERRAL_CODE = 0


def _reserve_to_dict(reserve: Any, reserve_keys: list[str]) -> dict[str, Any]:
    if isinstance(reserve, dict):
        return dict(reserve)
    return dict(zip(reserve_keys, reserve, strict=False))


def _compute_supply_cap_headroom(
    reserve: dict[str, Any], decimals: int
) -> tuple[int | None, int | None]:
    supply_cap_tokens = int(reserve.get("supplyCap") or 0)
    if supply_cap_tokens <= 0:
        return (None, None)
    unit = 10 ** max(0, int(decimals))
    supply_cap_wei = supply_cap_tokens * unit

    available = int(reserve.get("availableLiquidity") or 0)
    scaled_variable_debt = int(reserve.get("totalScaledVariableDebt") or 0)
    variable_index = int(reserve.get("variableBorrowIndex") or 0)
    current_variable_debt = (scaled_variable_debt * variable_index) // RAY

    # Note: stable debt is not included here because it is not exposed
    # via UI_POOL_RESERVE_KEYS / UI_POOL_DATA_PROVIDER_ABI for tuple-based data.
    total_supplied = available + current_variable_debt
    headroom = supply_cap_wei - total_supplied
    if headroom < 0:
        headroom = 0
    return (headroom, supply_cap_tokens)


class HyperlendAdapter(BaseAdapter):
    adapter_type = "HYPERLEND"

    def __init__(
        self,
        config: dict[str, Any] | None = None,
        strategy_wallet_signing_callback=None,
    ) -> None:
        super().__init__("hyperlend_adapter", config)
        config = config or {}

        self.strategy_wallet_signing_callback = strategy_wallet_signing_callback

        self.ledger_adapter = LedgerAdapter()
        self.token_adapter = TokenAdapter()
        strategy_wallet = config.get("strategy_wallet") or {}
        strategy_addr = strategy_wallet.get("address")

        self.strategy_wallet_address: str | None = (
            to_checksum_address(strategy_addr) if strategy_addr else None
        )
        self._variable_debt_token_by_underlying: dict[str, str] = {}

    async def get_stable_markets(
        self,
        *,
        required_underlying_tokens: float | None = None,
        buffer_bps: int | None = None,
        min_buffer_tokens: float | None = None,
    ) -> tuple[bool, StableMarketsHeadroomResponse | str]:
        try:
            data = await HYPERLEND_CLIENT.get_stable_markets(
                required_underlying_tokens=required_underlying_tokens,
                buffer_bps=buffer_bps,
                min_buffer_tokens=min_buffer_tokens,
            )
            return True, data
        except Exception as exc:
            return False, str(exc)

    async def get_assets_view(
        self,
        *,
        user_address: str,
    ) -> tuple[bool, AssetsView | str]:
        try:
            data = await HYPERLEND_CLIENT.get_assets_view(user_address=user_address)
            return True, data
        except Exception as exc:
            return False, str(exc)

    async def get_full_user_state(
        self,
        *,
        account: str,
        include_zero_positions: bool = False,
    ) -> tuple[bool, dict[str, Any] | str]:
        ok, view = await self.get_assets_view(user_address=account)
        if not ok:
            return False, str(view)

        assets = view.get("assets", []) if isinstance(view, dict) else []
        if include_zero_positions:
            positions = assets
        else:
            positions = [
                a
                for a in assets
                if float(a.get("supply", 0) or 0) > 0
                or float(a.get("variable_borrow", 0) or 0) > 0
            ]

        return (
            True,
            {
                "protocol": "hyperlend",
                "account": account,
                "positions": positions,
                "accountData": view.get("account_data")
                if isinstance(view, dict)
                else {},
                "assetsView": view,
            },
        )

    async def get_market_entry(
        self,
        *,
        token: str,
    ) -> tuple[bool, MarketEntry | str]:
        try:
            data = await HYPERLEND_CLIENT.get_market_entry(token=token)
            return True, data
        except Exception as exc:
            return False, str(exc)

    async def get_lend_rate_history(
        self,
        *,
        token: str,
        lookback_hours: int,
        force_refresh: bool | None = None,
    ) -> tuple[bool, LendRateHistory | str]:
        try:
            data = await HYPERLEND_CLIENT.get_lend_rate_history(
                token=token,
                lookback_hours=lookback_hours,
                force_refresh=force_refresh,
            )
            return True, data
        except Exception as exc:
            return False, str(exc)

    async def get_all_markets(self) -> tuple[bool, list[dict[str, Any]] | str]:
        try:
            async with web3_from_chain_id(CHAIN_ID_HYPEREVM) as web3:
                ui_pool = web3.eth.contract(
                    address=HYPERLEND_UI_POOL_DATA_PROVIDER,
                    abi=UI_POOL_DATA_PROVIDER_ABI,
                )

                reserves, base_currency = await ui_pool.functions.getReservesData(
                    HYPERLEND_POOL_ADDRESSES_PROVIDER
                ).call(block_identifier="pending")

                try:
                    ref_unit = int(base_currency[0]) if base_currency else 1
                except (TypeError, ValueError):
                    ref_unit = 1
                if not ref_unit:
                    ref_unit = 1

                try:
                    ref_usd_raw = int(base_currency[1]) if base_currency else 0
                except (TypeError, ValueError):
                    ref_usd_raw = 0

                try:
                    ref_usd_decimals = int(base_currency[3]) if base_currency else 0
                except (TypeError, ValueError):
                    ref_usd_decimals = 0

                ref_usd = (
                    ref_usd_raw / (10**ref_usd_decimals)
                    if ref_usd_decimals and ref_usd_decimals > 0
                    else float(ref_usd_raw)
                )

                reserve_keys = UI_POOL_RESERVE_KEYS

                markets: list[dict[str, Any]] = []
                for reserve in reserves or []:
                    r = _reserve_to_dict(reserve, reserve_keys)

                    underlying = to_checksum_address(str(r.get("underlyingAsset")))
                    symbol_raw = r.get("symbol") or ""
                    decimals = int(r.get("decimals") or 18)
                    a_token = to_checksum_address(str(r.get("aTokenAddress")))
                    v_debt = to_checksum_address(str(r.get("variableDebtTokenAddress")))
                    self._variable_debt_token_by_underlying[underlying.lower()] = v_debt

                    is_active = bool(r.get("isActive"))
                    is_frozen = bool(r.get("isFrozen"))
                    is_paused = bool(r.get("isPaused"))
                    is_siloed = bool(r.get("isSiloedBorrowing"))

                    liquidity_rate_ray = int(r.get("liquidityRate") or 0)
                    variable_borrow_rate_ray = int(r.get("variableBorrowRate") or 0)

                    price_market_ref = int(r.get("priceInMarketReferenceCurrency") or 0)
                    try:
                        price_market_ref_float = (
                            float(price_market_ref) / ref_unit if ref_unit else 0.0
                        )
                    except ZeroDivisionError:
                        price_market_ref_float = 0.0
                    price_usd = price_market_ref_float * ref_usd

                    supply_apr = ray_to_apr(liquidity_rate_ray)
                    borrow_apr = ray_to_apr(variable_borrow_rate_ray)

                    available_liquidity = int(r.get("availableLiquidity") or 0)
                    scaled_variable_debt = int(r.get("totalScaledVariableDebt") or 0)
                    variable_index = int(r.get("variableBorrowIndex") or 0)
                    total_variable_debt = (scaled_variable_debt * variable_index) // RAY
                    tvl = available_liquidity + total_variable_debt

                    symbol_canonical = normalize_symbol(symbol_raw) or normalize_symbol(
                        underlying
                    )

                    headroom_wei, supply_cap_tokens = _compute_supply_cap_headroom(
                        r, decimals
                    )

                    markets.append(
                        {
                            "underlying": underlying,
                            "symbol": str(symbol_raw),
                            "symbol_canonical": symbol_canonical,
                            "decimals": int(decimals),
                            "a_token": a_token,
                            "variable_debt_token": v_debt,
                            "is_active": is_active,
                            "is_frozen": is_frozen,
                            "is_paused": is_paused,
                            "is_siloed_borrowing": is_siloed,
                            "is_stablecoin": is_stable_symbol(symbol_raw),
                            "usage_as_collateral_enabled": bool(
                                r.get("usageAsCollateralEnabled")
                            ),
                            "ltv_bps": int(r.get("baseLTVasCollateral") or 0),
                            "liquidation_threshold_bps": int(
                                r.get("reserveLiquidationThreshold") or 0
                            ),
                            "liquidation_bonus_bps": int(
                                r.get("reserveLiquidationBonus") or 0
                            ),
                            "reserve_factor_bps": int(r.get("reserveFactor") or 0),
                            "borrowing_enabled": bool(r.get("borrowingEnabled")),
                            "borrow_cap": int(r.get("borrowCap") or 0),
                            "price_usd": float(price_usd),
                            "supply_apr": float(supply_apr),
                            "supply_apy": float(apr_to_apy(supply_apr)),
                            "variable_borrow_apr": float(borrow_apr),
                            "variable_borrow_apy": float(apr_to_apy(borrow_apr)),
                            "available_liquidity": int(available_liquidity),
                            "total_variable_debt": int(total_variable_debt),
                            "tvl": int(tvl),
                            "supply_cap": supply_cap_tokens,
                            "supply_cap_headroom": headroom_wei,
                            "debt_ceiling": int(r.get("debtCeiling") or 0),
                            "debt_ceiling_decimals": int(
                                r.get("debtCeilingDecimals") or 0
                            ),
                        }
                    )

                return True, markets
        except Exception as exc:  # noqa: BLE001
            return False, str(exc)

    async def lend(
        self,
        *,
        underlying_token: str,
        qty: int,
        chain_id: int,
        native: bool = False,
        strategy_name: str | None = None,
    ) -> tuple[bool, Any]:
        strategy = self.strategy_wallet_address
        if not strategy:
            return False, "strategy wallet address not configured"
        if qty <= 0:
            return False, "qty must be positive"

        if native:
            token_addr = HYPEREVM_WHYPE
            transaction = await encode_call(
                target=HYPERLEND_WRAPPED_TOKEN_GATEWAY,
                abi=WRAPPED_TOKEN_GATEWAY_ABI,
                fn_name="depositETH",
                args=[HYPEREVM_WHYPE, strategy, REFERRAL_CODE],
                from_address=strategy,
                chain_id=chain_id,
                value=qty,
            )
        else:
            token_addr = to_checksum_address(underlying_token)
            approved = await ensure_allowance(
                token_address=token_addr,
                owner=strategy,
                spender=HYPERLEND_POOL,
                amount=qty,
                chain_id=chain_id,
                signing_callback=self.strategy_wallet_signing_callback,
            )
            if not approved[0]:
                return approved
            transaction = await encode_call(
                target=HYPERLEND_POOL,
                abi=POOL_ABI,
                fn_name="supply",
                args=[token_addr, qty, strategy, REFERRAL_CODE],
                from_address=strategy,
                chain_id=chain_id,
            )

        txn_hash = await send_transaction(
            transaction, self.strategy_wallet_signing_callback
        )

        await self._record_pool_op(
            token_address=token_addr,
            amount=qty,
            chain_id=chain_id,
            wallet_address=strategy,
            txn_hash=txn_hash,
            strategy_name=strategy_name,
            op_type="lend",
        )

        return (True, txn_hash)

    async def unlend(
        self,
        *,
        underlying_token: str,
        qty: int,
        chain_id: int,
        native: bool = False,
        strategy_name: str | None = None,
    ) -> tuple[bool, Any]:
        strategy = self.strategy_wallet_address
        if not strategy:
            return False, "strategy wallet address not configured"
        if qty <= 0:
            return False, "qty must be positive"

        if native:
            token_addr = HYPEREVM_WHYPE
            transaction = await encode_call(
                target=HYPERLEND_WRAPPED_TOKEN_GATEWAY,
                abi=WRAPPED_TOKEN_GATEWAY_ABI,
                fn_name="withdrawETH",
                args=[HYPEREVM_WHYPE, qty, strategy],
                from_address=strategy,
                chain_id=chain_id,
            )
        else:
            token_addr = to_checksum_address(underlying_token)
            transaction = await encode_call(
                target=HYPERLEND_POOL,
                abi=POOL_ABI,
                fn_name="withdraw",
                args=[token_addr, qty, strategy],
                from_address=strategy,
                chain_id=chain_id,
            )

        txn_hash = await send_transaction(
            transaction, self.strategy_wallet_signing_callback
        )
        await self._record_pool_op(
            token_address=token_addr,
            amount=qty,
            chain_id=chain_id,
            wallet_address=strategy,
            txn_hash=txn_hash,
            strategy_name=strategy_name,
            op_type="unlend",
        )

        return True, txn_hash

    async def borrow(
        self,
        *,
        underlying_token: str,
        qty: int,
        chain_id: int,
        native: bool = False,
    ) -> tuple[bool, Any]:
        strategy = self.strategy_wallet_address
        if not strategy:
            return False, "strategy wallet address not configured"
        if qty <= 0:
            return False, "qty must be positive"

        if native:
            asset = HYPEREVM_WHYPE
            borrow_tx = await encode_call(
                target=HYPERLEND_POOL,
                abi=POOL_ABI,
                fn_name="borrow",
                args=[asset, qty, VARIABLE_RATE_MODE, REFERRAL_CODE, strategy],
                from_address=strategy,
                chain_id=chain_id,
            )
            borrow_tx_hash = await send_transaction(
                borrow_tx, self.strategy_wallet_signing_callback
            )

            unwrap_tx = await encode_call(
                target=asset,
                abi=WETH_ABI,
                fn_name="withdraw",
                args=[qty],
                from_address=strategy,
                chain_id=chain_id,
            )
            unwrap_tx_hash = await send_transaction(
                unwrap_tx, self.strategy_wallet_signing_callback
            )
            return True, {"borrow_tx": borrow_tx_hash, "unwrap_tx": unwrap_tx_hash}
        else:
            asset = to_checksum_address(underlying_token)
            transaction = await encode_call(
                target=HYPERLEND_POOL,
                abi=POOL_ABI,
                fn_name="borrow",
                args=[asset, qty, VARIABLE_RATE_MODE, REFERRAL_CODE, strategy],
                from_address=strategy,
                chain_id=chain_id,
            )

        txn_hash = await send_transaction(
            transaction, self.strategy_wallet_signing_callback
        )
        return True, txn_hash

    async def repay(
        self,
        *,
        underlying_token: str,
        qty: int,
        chain_id: int,
        native: bool = False,
        repay_full: bool = False,
    ) -> tuple[bool, Any]:
        strategy = self.strategy_wallet_address
        if not strategy:
            return False, "strategy wallet address not configured"
        if qty <= 0 and not repay_full:
            return False, "qty must be positive"

        if native:
            if repay_full:
                async with web3_from_chain_id(chain_id) as web3:
                    cache_key = HYPEREVM_WHYPE.lower()
                    variable_debt_token = self._variable_debt_token_by_underlying.get(
                        cache_key
                    )
                    if not variable_debt_token:
                        ui_pool = web3.eth.contract(
                            address=HYPERLEND_UI_POOL_DATA_PROVIDER,
                            abi=UI_POOL_DATA_PROVIDER_ABI,
                        )
                        reserves, _ = await ui_pool.functions.getReservesData(
                            HYPERLEND_POOL_ADDRESSES_PROVIDER
                        ).call(block_identifier="pending")

                        reserve_keys = UI_POOL_RESERVE_KEYS
                        for reserve in reserves or []:
                            r = _reserve_to_dict(reserve, reserve_keys)
                            underlying = str(r.get("underlyingAsset") or "")
                            if underlying and underlying.lower() == cache_key:
                                v = r.get("variableDebtTokenAddress")
                                if v:
                                    variable_debt_token = to_checksum_address(str(v))
                                break

                        if not variable_debt_token:
                            return (
                                False,
                                "could not resolve variable debt token for WHYPE",
                            )

                        self._variable_debt_token_by_underlying[cache_key] = (
                            variable_debt_token
                        )

                    variable_debt = await get_token_balance(
                        str(variable_debt_token),
                        chain_id,
                        strategy,
                        web3=web3,
                        block_identifier="pending",
                    )
                    if variable_debt <= 0:
                        return True, None

                    native_balance = await get_token_balance(
                        None,
                        chain_id,
                        strategy,
                        web3=web3,
                        block_identifier="pending",
                    )

                    # Send a small buffer to avoid leaving dust from interest accrual between
                    # the read and execution; excess is expected to be refunded by the gateway.
                    buffer_wei = max(1, variable_debt // 10_000)  # 0.01%
                    value = variable_debt + buffer_wei
                    if native_balance < value:
                        if native_balance < variable_debt:
                            return (
                                False,
                                f"insufficient HYPE balance for repay_full (debt_wei={variable_debt}, balance_wei={native_balance})",
                            )
                        value = variable_debt

                transaction = await encode_call(
                    target=HYPERLEND_WRAPPED_TOKEN_GATEWAY,
                    abi=WRAPPED_TOKEN_GATEWAY_ABI,
                    fn_name="repayETH",
                    args=[HYPEREVM_WHYPE, MAX_UINT256, strategy],
                    from_address=strategy,
                    chain_id=chain_id,
                    value=value,
                )
            else:
                transaction = await encode_call(
                    target=HYPERLEND_WRAPPED_TOKEN_GATEWAY,
                    abi=WRAPPED_TOKEN_GATEWAY_ABI,
                    fn_name="repayETH",
                    args=[HYPEREVM_WHYPE, qty, strategy],
                    from_address=strategy,
                    chain_id=chain_id,
                    value=qty,
                )

        else:
            repay_amount = MAX_UINT256 if repay_full else qty
            asset = to_checksum_address(underlying_token)
            allowance_target = MAX_UINT256 if repay_full else qty
            approved = await ensure_allowance(
                token_address=asset,
                owner=strategy,
                spender=HYPERLEND_POOL,
                amount=allowance_target,
                chain_id=chain_id,
                signing_callback=self.strategy_wallet_signing_callback,
                approval_amount=MAX_UINT256,
            )
            if not approved[0]:
                return approved

            transaction = await encode_call(
                target=HYPERLEND_POOL,
                abi=POOL_ABI,
                fn_name="repay",
                args=[asset, repay_amount, VARIABLE_RATE_MODE, strategy],
                from_address=strategy,
                chain_id=chain_id,
            )

        txn_hash = await send_transaction(
            transaction, self.strategy_wallet_signing_callback
        )
        return True, txn_hash

    async def set_collateral(
        self,
        *,
        underlying_token: str,
        chain_id: int,
    ) -> tuple[bool, Any]:
        strategy = self.strategy_wallet_address
        if not strategy:
            return False, "strategy wallet address not configured"

        asset = to_checksum_address(underlying_token)
        transaction = await encode_call(
            target=HYPERLEND_POOL,
            abi=POOL_ABI,
            fn_name="setUserUseReserveAsCollateral",
            args=[asset, True],
            from_address=strategy,
            chain_id=chain_id,
        )
        txn_hash = await send_transaction(
            transaction, self.strategy_wallet_signing_callback
        )
        return True, txn_hash

    async def remove_collateral(
        self,
        *,
        underlying_token: str,
        chain_id: int,
    ) -> tuple[bool, Any]:
        strategy = self.strategy_wallet_address
        if not strategy:
            return False, "strategy wallet address not configured"

        asset = to_checksum_address(underlying_token)
        transaction = await encode_call(
            target=HYPERLEND_POOL,
            abi=POOL_ABI,
            fn_name="setUserUseReserveAsCollateral",
            args=[asset, False],
            from_address=strategy,
            chain_id=chain_id,
        )
        txn_hash = await send_transaction(
            transaction, self.strategy_wallet_signing_callback
        )
        return True, txn_hash

    async def _record_pool_op(
        self,
        token_address: str,
        amount: int,
        chain_id: int,
        wallet_address: str,
        txn_hash: str,
        op_type: Literal["lend", "unlend"],
        strategy_name: str | None = None,
    ):
        amount_usd = await self._calculate_amount_usd(
            token_address=token_address,
            amount=amount,
            chain_id=chain_id,
        )

        model = {"lend": LEND, "unlend": UNLEND}[op_type]

        operation_data = model(
            adapter=self.adapter_type,
            token_address=token_address,
            pool_address=HYPERLEND_POOL,
            amount=str(amount),
            amount_usd=amount_usd or 0,
            transaction_hash=txn_hash,
            transaction_chain_id=chain_id,
        )

        success, ledger_response = await self.ledger_adapter.record_operation(
            wallet_address=wallet_address,
            operation_data=operation_data,
            usd_value=amount_usd or 0,
            strategy_name=strategy_name,
        )
        if not success:
            self.logger.warning("Ledger record failed", error=ledger_response)

    async def _calculate_amount_usd(
        self,
        token_address: str,
        amount: int,
        chain_id: int,
    ) -> float | None:
        success, token_data = await self.token_adapter.get_token(
            query=token_address,
            chain_id=chain_id,
        )
        if not success or not token_data:
            self.logger.warning(
                f"Could not get token info for {token_address} on chain {chain_id}"
            )
            return None

        decimals, current_price = (
            token_data["decimals"],
            token_data["current_price"],
        )

        if decimals is None or current_price is None:
            self.logger.warning(
                f"Could not get decimal or current_price info for {token_address} on chain {chain_id}"
            )
            return None

        return float(current_price) * float(amount) / 10 ** int(decimals)



from collections.abc import Awaitable, Callable
from typing import Any

from eth_account import Account
from eth_account.signers.local import LocalAccount


def _resolve_private_key(config: dict[str, Any]) -> str | None:
    """Extract private key from config."""
    strategy_wallet = config.get("strategy_wallet", {})
    if isinstance(strategy_wallet, dict):
        pk = strategy_wallet.get("private_key_hex") or strategy_wallet.get(
            "private_key"
        )
        if pk:
            return pk

    main_wallet = config.get("main_wallet", {})
    if isinstance(main_wallet, dict):
        pk = main_wallet.get("private_key_hex") or main_wallet.get("private_key")
        if pk:
            return pk

    return None


def create_local_signer(
    config: dict[str, Any],
) -> Callable[[dict[str, Any], str, str], Awaitable[dict[str, str] | None]]:
    private_key = _resolve_private_key(config)
    if not private_key:
        raise ValueError(
            "No private key found in config. "
            "Provide strategy_wallet.private_key_hex or strategy_wallet.private_key"
        )

    pk = private_key if private_key.startswith("0x") else "0x" + private_key
    account: LocalAccount = Account.from_key(pk)

    async def sign(
        action: dict[str, Any], payload: str, address: str
    ) -> dict[str, str] | None:
        if address.lower() != account.address.lower():
            return None

        signed = account.sign_message(payload)
        return {"r": hex(signed.r), "s": hex(signed.s), "v": signed.v}

    return sign



from hyperliquid.info import Info
from hyperliquid.utils import constants

_INFO: Info | None = None
_PERP_DEXES: list[str] = [""]


def get_info() -> Info:
    global _INFO, _PERP_DEXES
    if _INFO is None:
        info = Info(constants.MAINNET_API_URL, skip_ws=True)
        _PERP_DEXES = [""] + [i["name"] for i in info.perp_dexs() if i is not None]
        _INFO = Info(constants.MAINNET_API_URL, skip_ws=True, perp_dexs=_PERP_DEXES)
    return _INFO


def get_perp_dexes() -> list[str]:
    get_info()  # ensure initialized
    return _PERP_DEXES



from unittest.mock import AsyncMock, patch

import pytest

from wayfinder_paths.adapters.token_adapter.adapter import TokenAdapter


class TestTokenAdapter:
    @pytest.fixture
    def mock_token_client(self):
        return AsyncMock()

    @pytest.fixture
    def adapter(self):
        return TokenAdapter()

    def test_init_with_default_config(self):
        adapter = TokenAdapter()
        assert adapter.adapter_type == "TOKEN"

    @pytest.mark.asyncio
    async def test_get_token_success(self, adapter, mock_token_client):
        mock_token_data = {
            "address": "0x1234...",
            "symbol": "TEST",
            "name": "Test Token",
            "decimals": 18,
        }
        mock_token_client.get_token_details = AsyncMock(return_value=mock_token_data)

        with patch(
            "wayfinder_paths.adapters.token_adapter.adapter.TOKEN_CLIENT",
            mock_token_client,
        ):
            success, data = await adapter.get_token("0x1234...")

            assert success
            assert data == mock_token_data

    @pytest.mark.asyncio
    async def test_get_token_not_found(self, adapter, mock_token_client):
        mock_token_client.get_token_details = AsyncMock(return_value=None)

        with patch(
            "wayfinder_paths.adapters.token_adapter.adapter.TOKEN_CLIENT",
            mock_token_client,
        ):
            success, data = await adapter.get_token("0x1234...")

            assert success is False
            assert "No token found for" in data

    @pytest.mark.asyncio
    async def test_get_token_falls_back_to_onchain(self, adapter, mock_token_client):
        token_address = "0x" + ("1" * 40)
        mock_token_client.get_token_details = AsyncMock(return_value=None)

        onchain_data = {
            "address": token_address,
            "symbol": "TEST",
            "name": "Test Token",
            "decimals": 6,
        }

        with (
            patch(
                "wayfinder_paths.adapters.token_adapter.adapter.TOKEN_CLIENT",
                mock_token_client,
            ),
            patch.object(
                adapter,
                "get_token_onchain",
                AsyncMock(return_value=(True, onchain_data)),
            ),
        ):
            success, data = await adapter.get_token(token_address, chain_id=8453)

            assert success
            assert data == onchain_data

    @pytest.mark.asyncio
    async def test_get_token_by_token_id(self, adapter, mock_token_client):
        mock_token_data = {"address": "0x1234...", "symbol": "TEST"}
        mock_token_client.get_token_details = AsyncMock(return_value=mock_token_data)

        with patch(
            "wayfinder_paths.adapters.token_adapter.adapter.TOKEN_CLIENT",
            mock_token_client,
        ):
            success, data = await adapter.get_token("token-123")

            assert success
            assert data == mock_token_data

    def test_adapter_type(self, adapter):
        assert adapter.adapter_type == "TOKEN"

    @pytest.mark.asyncio
    async def test_get_token_price_success(self, adapter, mock_token_client):
        mock_token_data = {
            "current_price": 1.50,
            "price_change_24h": 0.05,
            "market_cap": 1000000,
            "total_volume_usd_24h": 50000,
            "symbol": "TEST",
            "name": "Test Token",
            "address": "0x1234...",
        }
        mock_token_client.get_token_details = AsyncMock(return_value=mock_token_data)

        with patch(
            "wayfinder_paths.adapters.token_adapter.adapter.TOKEN_CLIENT",
            mock_token_client,
        ):
            success, data = await adapter.get_token_price("test-token")

            assert success
            assert data["current_price"] == 1.50
            assert data["symbol"] == "TEST"
            assert data["name"] == "Test Token"
            assert data["total_volume"] == 50000
            assert data["price_change_percentage_24h"] == 5.0

    @pytest.mark.asyncio
    async def test_get_token_price_not_found(self, adapter, mock_token_client):
        mock_token_client.get_token_details = AsyncMock(return_value=None)

        with patch(
            "wayfinder_paths.adapters.token_adapter.adapter.TOKEN_CLIENT",
            mock_token_client,
        ):
            success, data = await adapter.get_token_price("invalid-token")

            assert success is False
            assert "No token found for" in data

    @pytest.mark.asyncio
    async def test_get_gas_token_success(self, adapter, mock_token_client):
        mock_gas_token_data = {
            "id": "ethereum_0x0000000000000000000000000000000000000000",
            "token_id": "ethereum_0x0000000000000000000000000000000000000000",
            "symbol": "ETH",
            "name": "Ethereum",
            "address": "0x0000000000000000000000000000000000000000",
            "decimals": 18,
        }
        mock_token_client.get_gas_token = AsyncMock(return_value=mock_gas_token_data)

        with patch(
            "wayfinder_paths.adapters.token_adapter.adapter.TOKEN_CLIENT",
            mock_token_client,
        ):
            success, data = await adapter.get_gas_token("base")

            assert success
            assert data["symbol"] == "ETH"
            assert data["name"] == "Ethereum"

    @pytest.mark.asyncio
    async def test_get_gas_token_not_found(self, adapter, mock_token_client):
        mock_token_client.get_gas_token = AsyncMock(return_value=None)

        with patch(
            "wayfinder_paths.adapters.token_adapter.adapter.TOKEN_CLIENT",
            mock_token_client,
        ):
            success, data = await adapter.get_gas_token("invalid-chain")

            assert success is False
            assert "No gas token found for chain" in data



from __future__ import annotations

import pytest

from wayfinder_paths.core.utils.wallets import (
    load_wallets,
    make_wallet_from_mnemonic,
    write_wallet_to_json,
)


def test_write_wallet_to_json_is_idempotent_for_same_wallet(tmp_path) -> None:
    mnemonic = "test test test test test test test test test test test junk"
    w = make_wallet_from_mnemonic(mnemonic, account_index=0)
    w["label"] = "main"

    write_wallet_to_json(w, out_dir=tmp_path, filename="config.json")
    write_wallet_to_json(w, out_dir=tmp_path, filename="config.json")

    wallets = load_wallets(tmp_path, "config.json")
    assert len(wallets) == 1
    assert wallets[0]["address"].lower() == w["address"].lower()
    assert wallets[0]["label"] == "main"


def test_write_wallet_to_json_refuses_overwrite_by_address(tmp_path) -> None:
    mnemonic = "test test test test test test test test test test test junk"
    w = make_wallet_from_mnemonic(mnemonic, account_index=0)
    w["label"] = "main"
    write_wallet_to_json(w, out_dir=tmp_path, filename="config.json")

    w2 = dict(w)
    w2["label"] = "other"
    with pytest.raises(ValueError, match="refusing to overwrite"):
        write_wallet_to_json(w2, out_dir=tmp_path, filename="config.json")


def test_write_wallet_to_json_refuses_duplicate_label(tmp_path) -> None:
    mnemonic = "test test test test test test test test test test test junk"
    w0 = make_wallet_from_mnemonic(mnemonic, account_index=0)
    w0["label"] = "main"
    write_wallet_to_json(w0, out_dir=tmp_path, filename="config.json")

    w1 = make_wallet_from_mnemonic(mnemonic, account_index=1)
    w1["label"] = "main"
    with pytest.raises(ValueError, match="refusing to create duplicate"):
        write_wallet_to_json(w1, out_dir=tmp_path, filename="config.json")



"""Boros-specific utilities (tick math, parsing helpers, conversions)."""

from __future__ import annotations

import math
from decimal import Decimal
from typing import Any

BOROS_TICK_BASE = 1.0001


def tick_from_rate(
    rate: float, tick_step: int, *, round_down: bool, base: float = BOROS_TICK_BASE
) -> int:
    """Convert APR rate to Boros limitTick."""
    if tick_step <= 0:
        tick_step = 1
    ln_base = math.log(base)
    if rate >= 0:
        if rate == 0:
            return 0
        raw = math.log1p(rate) / (tick_step * ln_base)
        return int(math.floor(raw) if round_down else math.ceil(raw))
    # Negative rate
    raw = math.log1p(-rate) / (tick_step * ln_base)
    return -int(math.floor(raw) if round_down else math.ceil(raw))


def rate_from_tick(tick: int, tick_step: int, base: float = BOROS_TICK_BASE) -> float:
    """Convert Boros limitTick to APR rate."""
    if tick_step <= 0:
        tick_step = 1
    p = base ** (abs(int(tick)) * int(tick_step))
    r = p - 1
    return r if tick >= 0 else -r


def normalize_apr(value: Any) -> float | None:
    """Normalize various APR encodings to decimal.

    Handles: decimal (0.1115), percent (11.15), bps (1115), 1e18-scaled.
    """
    if value is None:
        return None
    try:
        x = float(value)
    except (TypeError, ValueError):
        return None

    if x == 0:
        return None
    # 1e18-scaled decimal
    if x > 1e9:
        return x / 1e18
    # bps (1115 = 11.15%)
    if x > 1000:
        return x / 10_000.0
    # percent (11.15 = 11.15%)
    if x > 1:
        return x / 100.0
    # already decimal
    return x


def cash_wei_to_float(value_wei: Any) -> float:
    """Convert Boros cash units (1e18) to float."""
    if value_wei is None:
        return 0.0
    try:
        return float(Decimal(str(value_wei)) / Decimal(1e18))
    except Exception:
        return 0.0


def market_id_from_market_acc(market_acc: str) -> int | None:
    """Parse a Boros `marketAcc` into a market_id (last 3 bytes)."""
    if not market_acc or len(market_acc) < 8:
        return None
    try:
        market_id = int(market_acc[-6:], 16)
    except ValueError:
        return None
    if market_id == 0xFFFFFF:
        return None
    return market_id



# Minimal ABIs for Uniswap V3 NonfungiblePositionManager + Factory.
# Shared by all V3 fork adapters (Uniswap, ProjectX, etc.).

NONFUNGIBLE_POSITION_MANAGER_ABI = [
    {
        "name": "balanceOf",
        "type": "function",
        "stateMutability": "view",
        "inputs": [{"name": "owner", "type": "address"}],
        "outputs": [{"type": "uint256"}],
    },
    {
        "name": "tokenOfOwnerByIndex",
        "type": "function",
        "stateMutability": "view",
        "inputs": [
            {"name": "owner", "type": "address"},
            {"name": "index", "type": "uint256"},
        ],
        "outputs": [{"type": "uint256"}],
    },
    {
        "name": "positions",
        "type": "function",
        "stateMutability": "view",
        "inputs": [{"name": "tokenId", "type": "uint256"}],
        "outputs": [
            {"name": "nonce", "type": "uint96"},
            {"name": "operator", "type": "address"},
            {"name": "token0", "type": "address"},
            {"name": "token1", "type": "address"},
            {"name": "fee", "type": "uint24"},
            {"name": "tickLower", "type": "int24"},
            {"name": "tickUpper", "type": "int24"},
            {"name": "liquidity", "type": "uint128"},
            {"name": "feeGrowthInside0LastX128", "type": "uint256"},
            {"name": "feeGrowthInside1LastX128", "type": "uint256"},
            {"name": "tokensOwed0", "type": "uint128"},
            {"name": "tokensOwed1", "type": "uint128"},
        ],
    },
    {
        "name": "mint",
        "type": "function",
        "stateMutability": "payable",
        "inputs": [
            {
                "name": "params",
                "type": "tuple",
                "components": [
                    {"name": "token0", "type": "address"},
                    {"name": "token1", "type": "address"},
                    {"name": "fee", "type": "uint24"},
                    {"name": "tickLower", "type": "int24"},
                    {"name": "tickUpper", "type": "int24"},
                    {"name": "amount0Desired", "type": "uint256"},
                    {"name": "amount1Desired", "type": "uint256"},
                    {"name": "amount0Min", "type": "uint256"},
                    {"name": "amount1Min", "type": "uint256"},
                    {"name": "recipient", "type": "address"},
                    {"name": "deadline", "type": "uint256"},
                ],
            }
        ],
        "outputs": [
            {"name": "tokenId", "type": "uint256"},
            {"name": "liquidity", "type": "uint128"},
            {"name": "amount0", "type": "uint256"},
            {"name": "amount1", "type": "uint256"},
        ],
    },
    {
        "name": "increaseLiquidity",
        "type": "function",
        "stateMutability": "payable",
        "inputs": [
            {
                "name": "params",
                "type": "tuple",
                "components": [
                    {"name": "tokenId", "type": "uint256"},
                    {"name": "amount0Desired", "type": "uint256"},
                    {"name": "amount1Desired", "type": "uint256"},
                    {"name": "amount0Min", "type": "uint256"},
                    {"name": "amount1Min", "type": "uint256"},
                    {"name": "deadline", "type": "uint256"},
                ],
            }
        ],
        "outputs": [
            {"name": "liquidity", "type": "uint128"},
            {"name": "amount0", "type": "uint256"},
            {"name": "amount1", "type": "uint256"},
        ],
    },
    {
        "name": "decreaseLiquidity",
        "type": "function",
        "stateMutability": "nonpayable",
        "inputs": [
            {
                "name": "params",
                "type": "tuple",
                "components": [
                    {"name": "tokenId", "type": "uint256"},
                    {"name": "liquidity", "type": "uint128"},
                    {"name": "amount0Min", "type": "uint256"},
                    {"name": "amount1Min", "type": "uint256"},
                    {"name": "deadline", "type": "uint256"},
                ],
            }
        ],
        "outputs": [
            {"name": "amount0", "type": "uint256"},
            {"name": "amount1", "type": "uint256"},
        ],
    },
    {
        "name": "collect",
        "type": "function",
        "stateMutability": "payable",
        "inputs": [
            {
                "name": "params",
                "type": "tuple",
                "components": [
                    {"name": "tokenId", "type": "uint256"},
                    {"name": "recipient", "type": "address"},
                    {"name": "amount0Max", "type": "uint128"},
                    {"name": "amount1Max", "type": "uint128"},
                ],
            }
        ],
        "outputs": [
            {"name": "amount0", "type": "uint256"},
            {"name": "amount1", "type": "uint256"},
        ],
    },
    {
        "name": "burn",
        "type": "function",
        "stateMutability": "nonpayable",
        "inputs": [{"name": "tokenId", "type": "uint256"}],
        "outputs": [],
    },
    {
        "name": "multicall",
        "type": "function",
        "stateMutability": "payable",
        "inputs": [{"name": "data", "type": "bytes[]"}],
        "outputs": [{"name": "results", "type": "bytes[]"}],
    },
]

UNISWAP_V3_POOL_ABI = [
    {
        "name": "slot0",
        "type": "function",
        "stateMutability": "view",
        "inputs": [],
        "outputs": [
            {"name": "sqrtPriceX96", "type": "uint160"},
            {"name": "tick", "type": "int24"},
            {"name": "observationIndex", "type": "uint16"},
            {"name": "observationCardinality", "type": "uint16"},
            {"name": "observationCardinalityNext", "type": "uint16"},
            {"name": "feeProtocol", "type": "uint8"},
            {"name": "unlocked", "type": "bool"},
        ],
    },
]

UNISWAP_V3_FACTORY_ABI = [
    {
        "name": "getPool",
        "type": "function",
        "stateMutability": "view",
        "inputs": [
            {"name": "tokenA", "type": "address"},
            {"name": "tokenB", "type": "address"},
            {"name": "fee", "type": "uint24"},
        ],
        "outputs": [{"name": "pool", "type": "address"}],
    }
]



from __future__ import annotations

from typing import Any

from wayfinder_paths.core.constants.erc1155_abi import ERC1155_APPROVAL_ABI

POLYMARKET_GAMMA_BASE_URL = "https://gamma-api.polymarket.com"
POLYMARKET_CLOB_BASE_URL = "https://clober2-630978094941.asia-northeast1.run.app"
POLYMARKET_DATA_BASE_URL = "https://data-api.polymarket.com"
POLYMARKET_BRIDGE_BASE_URL = "https://bridge.polymarket.com"

POLYGON_CHAIN_ID = 137

# Collateral
POLYGON_USDC_E_ADDRESS = "0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174"
POLYGON_USDC_ADDRESS = "0x3c499c542cEF5E3811e1192ce70d8cC03d5c3359"

# Polymarket contracts (CTF)
POLYMARKET_CONDITIONAL_TOKENS_ADDRESS = "0x4D97DCd97eC945f40cF65F87097ACe5EA0476045"

# Exchanges / operators that may require approvals depending on market type.
POLYMARKET_CTF_EXCHANGE_ADDRESS = "0x4bFb41d5B3570DeFd03C39a9A4D8dE6Bd8B8982E"
POLYMARKET_NEG_RISK_CTF_EXCHANGE_ADDRESS = "0xC5d563A36AE78145C45a50134d48A1215220f80a"
POLYMARKET_RISK_ADAPTER_EXCHANGE_ADDRESS = "0xd91E80cF2E7be2e162c6513ceD06f1dD0dA35296"

POLYMARKET_APPROVAL_TARGETS: list[str] = [
    POLYMARKET_CTF_EXCHANGE_ADDRESS,
    POLYMARKET_NEG_RISK_CTF_EXCHANGE_ADDRESS,
    POLYMARKET_RISK_ADAPTER_EXCHANGE_ADDRESS,
]

# Some NegRisk markets pay out an adapter "collateral" token which must be unwrapped.
POLYMARKET_ADAPTER_COLLATERAL_ADDRESS = "0x3A3BD7bb9528E159577F7C2e685CC81A765002E2"

MAX_UINT256 = (1 << 256) - 1
ZERO32_STR = "0x" + "00" * 32

CONDITIONAL_TOKENS_ABI: list[dict[str, Any]] = [
    *ERC1155_APPROVAL_ABI,
    {
        "type": "function",
        "stateMutability": "view",
        "name": "getCollectionId",
        "inputs": [
            {"name": "parentCollectionId", "type": "bytes32"},
            {"name": "conditionId", "type": "bytes32"},
            {"name": "indexSet", "type": "uint256"},
        ],
        "outputs": [{"name": "", "type": "bytes32"}],
    },
    {
        "type": "function",
        "stateMutability": "view",
        "name": "getPositionId",
        "inputs": [
            {"name": "collateralToken", "type": "address"},
            {"name": "collectionId", "type": "bytes32"},
        ],
        "outputs": [{"name": "", "type": "uint256"}],
    },
    {
        "type": "function",
        "stateMutability": "view",
        "name": "balanceOf",
        "inputs": [
            {"name": "account", "type": "address"},
            {"name": "id", "type": "uint256"},
        ],
        "outputs": [{"name": "", "type": "uint256"}],
    },
    {
        "type": "function",
        "stateMutability": "nonpayable",
        "name": "redeemPositions",
        "inputs": [
            {"name": "collateralToken", "type": "address"},
            {"name": "parentCollectionId", "type": "bytes32"},
            {"name": "conditionId", "type": "bytes32"},
            {"name": "indexSets", "type": "uint256[]"},
        ],
        "outputs": [],
    },
]

TOKEN_UNWRAP_ABI: list[dict[str, Any]] = [
    {
        "type": "function",
        "stateMutability": "nonpayable",
        "name": "unwrap",
        "inputs": [
            {"name": "to", "type": "address"},
            {"name": "amount", "type": "uint256"},
        ],
        "outputs": [],
    }
]



from __future__ import annotations

import sys
from pathlib import Path

from wayfinder_paths.runner.daemon import RunnerDaemon
from wayfinder_paths.runner.db import RunnerDB
from wayfinder_paths.runner.paths import RunnerPaths
from wayfinder_paths.runner.script_resolver import resolve_script_path


def _paths(tmp_path: Path) -> RunnerPaths:
    runner_dir = tmp_path / ".wayfinder" / "runner"
    return RunnerPaths(
        repo_root=tmp_path,
        runner_dir=runner_dir,
        db_path=runner_dir / "state.db",
        logs_dir=runner_dir / "logs",
        sock_path=runner_dir / "runner.sock",
    )


def test_resolve_script_path_only_allows_wayfinder_runs(tmp_path: Path) -> None:
    p = _paths(tmp_path)
    runs_dir = tmp_path / ".wayfinder_runs"
    runs_dir.mkdir(parents=True, exist_ok=True)
    script = runs_dir / "hello.py"
    script.write_text("print('hi')\n", encoding="utf-8")

    resolved = resolve_script_path(p, ".wayfinder_runs/hello.py")
    assert resolved.exists()
    assert resolved.name == "hello.py"

    outside = tmp_path / "nope.py"
    outside.write_text("print('no')\n", encoding="utf-8")
    try:
        resolve_script_path(p, "nope.py")
    except ValueError as exc:
        assert "local runs directory" in str(exc)
    else:
        raise AssertionError("Expected ValueError for script outside .wayfinder_runs")


def test_script_job_builds_worker_cmd(tmp_path: Path) -> None:
    p = _paths(tmp_path)
    runs_dir = tmp_path / ".wayfinder_runs"
    runs_dir.mkdir(parents=True, exist_ok=True)
    script = runs_dir / "hello.py"
    script.write_text("print('hi')\n", encoding="utf-8")

    daemon = RunnerDaemon(paths=p)
    cmd = daemon._build_worker_cmd(
        job={
            "type": "script",
            "payload": {
                "script_path": ".wayfinder_runs/hello.py",
                "args": ["--x", "1"],
            },
        }
    )
    assert cmd[0] == sys.executable
    assert cmd[1].endswith("hello.py")
    assert cmd[-2:] == ["--x", "1"]


def test_daemon_adds_script_job_with_relative_path(tmp_path: Path) -> None:
    p = _paths(tmp_path)
    runs_dir = tmp_path / ".wayfinder_runs"
    runs_dir.mkdir(parents=True, exist_ok=True)
    script = runs_dir / "hello.py"
    script.write_text("print('hi')\n", encoding="utf-8")

    daemon = RunnerDaemon(paths=p)
    resp = daemon.ctl_add_job(
        name="script-job",
        job_type="script",
        payload={"script_path": str(script), "args": []},
        interval_seconds=60,
    )
    assert resp["ok"] is True

    db = RunnerDB(p.db_path)
    jobs = db.list_jobs()
    job = next(j for j in jobs if j["name"] == "script-job")
    stored = str(job["payload"]["script_path"])
    assert stored.endswith("hello.py")
    assert not Path(stored).is_absolute()



from typing import Annotated, Any, Literal

from pydantic import BaseModel, Field


class OperationBase(BaseModel):
    # These are provided by adapters at execution time, but tests and callers may
    # construct operations without them (e.g., for bookkeeping payloads).
    adapter: str = "unknown"
    transaction_hash: str | None = None
    transaction_chain_id: int | None = None


class SWAP(OperationBase):
    type: Literal["SWAP"] = "SWAP"
    from_token_id: str
    to_token_id: str
    from_amount: str
    to_amount: str
    from_amount_usd: float
    to_amount_usd: float
    transaction_status: str | None = None
    transaction_receipt: dict[str, Any] | None = None


class LEND(OperationBase):
    type: Literal["LEND"] = "LEND"
    token_address: str
    pool_address: str
    amount: str
    amount_usd: float
    transaction_status: str | None = None
    transaction_receipt: dict[str, Any] | None = None


class UNLEND(OperationBase):
    type: Literal["UNLEND"] = "UNLEND"
    token_address: str
    pool_address: str
    amount: str
    amount_usd: float
    transaction_status: str | None = None
    transaction_receipt: dict[str, Any] | None = None


Operation = SWAP | LEND | UNLEND


class STRAT_OP(BaseModel):
    op_data: Annotated[Operation, Field(discriminator="type")]



import os
from unittest.mock import AsyncMock

import pytest

from wayfinder_paths.adapters.hyperliquid_adapter.adapter import HyperliquidAdapter

if os.getenv("RUN_HYPERLIQUID_LIVE_TESTS", "").lower() not in ("1", "true", "yes"):
    pytest.skip(
        "Hyperliquid live tests are disabled (set RUN_HYPERLIQUID_LIVE_TESTS=1 to enable).",
        allow_module_level=True,
    )


@pytest.fixture
def live_adapter():
    return HyperliquidAdapter(config={})


class TestAdapterUsesLiveMids:
    @pytest.mark.asyncio
    async def test_place_market_order_builds_ioc_limit(self, live_adapter):
        asset_id = live_adapter.coin_to_asset["HYPE"]

        adapter = HyperliquidAdapter(
            config={},
            sign_callback=AsyncMock(return_value="0x" + "00" * 65),
        )

        async def _no_broadcast(action, address):
            return {"status": "ok", "action": action}

        adapter._sign_and_broadcast_hypecore = _no_broadcast

        success, result = await adapter.place_market_order(
            asset_id=asset_id,
            is_buy=True,
            slippage=0.01,
            size=1.0,
            address="0x0000000000000000000000000000000000000000",
        )

        action = result["action"]
        assert action["type"] == "order"
        assert action["orders"][0]["a"] == asset_id
        assert action["orders"][0]["b"] is True

        ok, mids = await live_adapter.get_all_mid_prices()
        assert ok
        mid = mids["HYPE"]
        px = float(action["orders"][0]["p"])
        assert px >= mid * 0.999



from __future__ import annotations

from typing import Any


def dispatch(daemon, *, method: str, params: dict[str, Any]) -> dict[str, Any]:
    """Dispatch a control-plane method to the daemon implementation.

    This module is intentionally transport-agnostic (Unix socket today; HTTP later).
    """

    if method == "status":
        return daemon.ctl_status()
    if method == "shutdown":
        return daemon.ctl_shutdown()
    if method == "job_runs":
        return daemon.ctl_job_runs(
            name=params.get("name"),
            limit=params.get("limit"),
        )
    if method == "run_report":
        return daemon.ctl_run_report(
            run_id=params.get("run_id"),
            tail_bytes=params.get("tail_bytes"),
        )
    if method == "add_job":
        return daemon.ctl_add_job(
            name=params.get("name"),
            job_type=params.get("type"),
            payload=params.get("payload") or {},
            interval_seconds=params.get("interval_seconds"),
        )
    if method == "update_job":
        return daemon.ctl_update_job(
            name=params.get("name"),
            payload=params.get("payload"),
            interval_seconds=params.get("interval_seconds"),
        )
    if method == "pause_job":
        return daemon.ctl_pause_job(name=params.get("name"))
    if method == "resume_job":
        return daemon.ctl_resume_job(name=params.get("name"))
    if method == "run_once":
        return daemon.ctl_run_once(name=params.get("name"))
    if method == "delete_job":
        return daemon.ctl_delete_job(name=params.get("name"))

    return {"ok": False, "error": f"unknown_method: {method}"}



from __future__ import annotations

import hashlib
from pathlib import Path
from typing import Any

from wayfinder_paths.core.constants.hyperliquid import (
    ARBITRUM_USDC_TOKEN_ID,
    HYPE_FEE_WALLET,
    HYPERLIQUID_BRIDGE_ADDRESS,
)
from wayfinder_paths.core.constants.polymarket import (
    POLYGON_USDC_ADDRESS,
    POLYGON_USDC_E_ADDRESS,
)
from wayfinder_paths.mcp.utils import (
    find_wallet_by_label,
    normalize_address,
    read_text_excerpt,
    repo_root,
)


def build_execution_preview(tool_input: dict[str, Any]) -> dict[str, Any]:
    req = tool_input.get("request") if isinstance(tool_input, dict) else None
    if not isinstance(req, dict):
        return {
            "summary": "Execute request missing 'request' object.",
            "recipient_mismatch": False,
        }

    kind = str(req.get("kind") or "").strip()
    wallet_label = str(req.get("wallet_label") or "").strip()
    w = find_wallet_by_label(wallet_label) if wallet_label else None
    sender = normalize_address((w or {}).get("address")) if w else None

    recipient = normalize_address(req.get("recipient"))
    if kind == "swap":
        recipient = recipient or sender
        summary = (
            "EXECUTE swap\n"
            f"wallet_label: {wallet_label}\n"
            f"from_token: {req.get('from_token')}\n"
            f"to_token: {req.get('to_token')}\n"
            f"amount: {req.get('amount')}\n"
            f"slippage_bps: {req.get('slippage_bps')}\n"
            f"sender: {sender or '(unknown)'}\n"
            f"recipient: {recipient or '(unknown)'}"
        )
    elif kind == "hyperliquid_deposit":
        recipient = normalize_address(HYPERLIQUID_BRIDGE_ADDRESS)
        summary = (
            "EXECUTE hyperliquid_deposit (Bridge2)\n"
            f"wallet_label: {wallet_label}\n"
            f"token: {ARBITRUM_USDC_TOKEN_ID}\n"
            f"amount: {req.get('amount')}\n"
            "chain_id: 42161\n"
            f"sender: {sender or '(unknown)'}\n"
            f"recipient: {recipient or '(missing)'}"
        )
    elif kind == "send":
        summary = (
            "EXECUTE send\n"
            f"wallet_label: {wallet_label}\n"
            f"token: {req.get('token')}\n"
            f"amount: {req.get('amount')}\n"
            f"chain_id: {req.get('chain_id')}\n"
            f"sender: {sender or '(unknown)'}\n"
            f"recipient: {recipient or '(missing)'}"
        )
    else:
        summary = f"EXECUTE {kind or '(unknown kind)'}\nwallet_label: {wallet_label}"

    mismatch = bool(sender and recipient and sender.lower() != recipient.lower())
    if kind == "hyperliquid_deposit":
        mismatch = False  # deposit recipient is fixed; mismatch is expected
    return {"summary": summary, "recipient_mismatch": mismatch}


def build_run_script_preview(tool_input: dict[str, Any]) -> dict[str, Any]:
    ti = tool_input if isinstance(tool_input, dict) else {}
    path_raw = ti.get("script_path") or ti.get("path")
    args = ti.get("args") if isinstance(ti.get("args"), list) else []

    if not isinstance(path_raw, str) or not path_raw.strip():
        return {"summary": "RUN_SCRIPT missing script_path."}

    root = repo_root()
    p = Path(path_raw)
    if not p.is_absolute():
        p = root / p
    resolved = p.resolve(strict=False)

    rel = str(resolved)
    try:
        rel = str(resolved.relative_to(root))
    except Exception:
        pass

    sha = None
    try:
        if resolved.exists():
            sha = hashlib.sha256(resolved.read_bytes()).hexdigest()
    except Exception:
        sha = None

    excerpt = read_text_excerpt(resolved, max_chars=1200) if resolved.exists() else None

    summary = (
        "RUN_SCRIPT (executes local python)\n"
        f"script_path: {rel}\n"
        f"args: {args or []}\n"
        f"script_sha256: {(sha[:12] + '…') if sha else '(unavailable)'}"
    )
    if excerpt:
        summary += "\n\n" + excerpt
    else:
        summary += "\n\n(no script contents available)"

    return {"summary": summary}


def build_hyperliquid_execute_preview(tool_input: dict[str, Any]) -> dict[str, Any]:
    # hyperliquid_execute uses direct parameters, not a 'request' wrapper
    req = tool_input if isinstance(tool_input, dict) else {}
    if not req:
        return {"summary": "HYPERLIQUID_EXECUTE missing parameters."}

    action = str(req.get("action") or "").strip()
    wallet_label = str(req.get("wallet_label") or "").strip()
    w = find_wallet_by_label(wallet_label) if wallet_label else None
    sender = normalize_address((w or {}).get("address")) if w else None

    coin = req.get("coin")
    asset_id = req.get("asset_id")

    header = "HYPERLIQUID_EXECUTE\n"
    base = (
        f"action: {action or '(missing)'}\n"
        f"wallet_label: {wallet_label}\n"
        f"address: {sender or '(unknown)'}\n"
        f"coin: {coin}\n"
        f"asset_id: {asset_id}"
    )

    if action == "place_order":
        details = (
            "\n\nORDER\n"
            f"order_type: {req.get('order_type')}\n"
            f"is_buy: {req.get('is_buy')}\n"
            f"size: {req.get('size')}\n"
            f"usd_amount: {req.get('usd_amount')}\n"
            f"usd_amount_kind: {req.get('usd_amount_kind')}\n"
            f"price: {req.get('price')}\n"
            f"slippage: {req.get('slippage')}\n"
            f"reduce_only: {req.get('reduce_only')}\n"
            f"cloid: {req.get('cloid')}\n"
            f"leverage: {req.get('leverage')}\n"
            f"is_cross: {req.get('is_cross')}\n"
            f"builder_wallet: {HYPE_FEE_WALLET}\n"
            f"builder_fee_tenths_bp: {req.get('builder_fee_tenths_bp') or '(from config/default)'}"
        )
        return {"summary": header + base + details}

    if action == "cancel_order":
        details = (
            "\n\nCANCEL\n"
            f"order_id: {req.get('order_id')}\n"
            f"cancel_cloid: {req.get('cancel_cloid')}"
        )
        return {"summary": header + base + details}

    if action == "update_leverage":
        details = (
            "\n\nLEVERAGE\n"
            f"leverage: {req.get('leverage')}\n"
            f"is_cross: {req.get('is_cross')}"
        )
        return {"summary": header + base + details}

    if action == "withdraw":
        details = f"\n\nWITHDRAW\namount_usdc: {req.get('amount_usdc')}"
        return {"summary": header + base + details}

    if action == "spot_to_perp_transfer":
        details = f"\n\nTRANSFER SPOT → PERP\nusd_amount: {req.get('usd_amount')}"
        return {"summary": header + base + details}

    if action == "perp_to_spot_transfer":
        details = f"\n\nTRANSFER PERP → SPOT\nusd_amount: {req.get('usd_amount')}"
        return {"summary": header + base + details}

    return {"summary": header + base}


def build_polymarket_execute_preview(tool_input: dict[str, Any]) -> dict[str, Any]:
    req = tool_input if isinstance(tool_input, dict) else {}
    if not req:
        return {
            "summary": "POLYMARKET_EXECUTE missing parameters.",
            "recipient_mismatch": False,
        }

    action = str(req.get("action") or "").strip()
    wallet_label = str(req.get("wallet_label") or "").strip()
    w = find_wallet_by_label(wallet_label) if wallet_label else None
    sender = normalize_address((w or {}).get("address")) if w else None

    recipient = None
    if action == "bridge_deposit":
        recipient = normalize_address(req.get("recipient_address"))
    if action == "bridge_withdraw":
        recipient = normalize_address(req.get("recipient_addr"))

    mismatch = bool(sender and recipient and sender.lower() != recipient.lower())

    header = "POLYMARKET_EXECUTE\n"
    base = (
        f"action: {action or '(missing)'}\n"
        f"wallet_label: {wallet_label or '(missing)'}\n"
        f"address: {sender or '(unknown)'}"
    )

    if action == "bridge_deposit":
        details = (
            "\n\nCONVERT (token → USDC.e)\n"
            "route: BRAP swap preferred; Polymarket Bridge fallback\n"
            f"polymarket_collateral_usdce: {POLYGON_USDC_E_ADDRESS}\n"
            f"polygon_native_usdc: {POLYGON_USDC_ADDRESS}\n"
            f"from_chain_id: {req.get('from_chain_id')}\n"
            f"from_token_address: {req.get('from_token_address')}\n"
            f"amount: {req.get('amount')}\n"
            f"recipient_address: {req.get('recipient_address')}\n"
            "note: If you already have USDC.e on Polygon, you can trade without running bridge_deposit."
        )
        return {"summary": header + base + details, "recipient_mismatch": mismatch}

    if action == "bridge_withdraw":
        details = (
            "\n\nCONVERT (USDC.e → token)\n"
            "route: BRAP swap preferred; Polymarket Bridge fallback\n"
            f"polymarket_collateral_usdce: {POLYGON_USDC_E_ADDRESS}\n"
            f"polygon_native_usdc: {POLYGON_USDC_ADDRESS}\n"
            f"amount_usdce: {req.get('amount_usdce')}\n"
            f"to_chain_id: {req.get('to_chain_id')}\n"
            f"to_token_address: {req.get('to_token_address')}\n"
            f"recipient_addr: {req.get('recipient_addr')}"
        )
        return {"summary": header + base + details, "recipient_mismatch": mismatch}

    if action in {"buy", "sell", "close_position"}:
        details = (
            "\n\nTRADE\n"
            f"market_slug: {req.get('market_slug')}\n"
            f"outcome: {req.get('outcome')}\n"
            f"token_id: {req.get('token_id')}\n"
            f"amount_usdc: {req.get('amount_usdc')}\n"
            f"shares: {req.get('shares')}"
        )
        return {"summary": header + base + details, "recipient_mismatch": False}

    if action == "place_limit_order":
        details = (
            "\n\nLIMIT ORDER\n"
            f"token_id: {req.get('token_id')}\n"
            f"side: {req.get('side')}\n"
            f"price: {req.get('price')}\n"
            f"size: {req.get('size')}\n"
            f"post_only: {req.get('post_only')}"
        )
        return {"summary": header + base + details, "recipient_mismatch": False}

    if action == "cancel_order":
        details = f"\n\nCANCEL ORDER\norder_id: {req.get('order_id')}"
        return {"summary": header + base + details, "recipient_mismatch": False}

    if action == "redeem_positions":
        details = f"\n\nREDEEM\ncondition_id: {req.get('condition_id')}"
        return {"summary": header + base + details, "recipient_mismatch": False}

    return {"summary": header + base, "recipient_mismatch": mismatch}



from contextlib import asynccontextmanager
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from wayfinder_paths.adapters.hyperlend_adapter.adapter import HyperlendAdapter
from wayfinder_paths.core.constants.base import MAX_UINT256
from wayfinder_paths.core.constants.contracts import (
    HYPEREVM_WHYPE,
    HYPERLEND_POOL,
    HYPERLEND_WRAPPED_TOKEN_GATEWAY,
)
from wayfinder_paths.core.constants.hyperlend_abi import UI_POOL_RESERVE_KEYS


class TestHyperlendAdapter:
    @pytest.fixture
    def mock_hyperlend_client(self):
        return AsyncMock()

    @pytest.fixture
    def adapter(self):
        return HyperlendAdapter(
            config={
                "strategy_wallet": {
                    "address": "0x1234567890123456789012345678901234567890"
                }
            },
        )

    @pytest.mark.asyncio
    async def test_get_stable_markets_success(self, adapter, mock_hyperlend_client):
        mock_response = {
            "markets": {
                "0x1234...": {
                    "symbol": "USDT",
                    "symbol_canonical": "usdt",
                    "display_symbol": "USDT",
                    "reserve": {},
                    "decimals": 6,
                    "headroom": 1000000000000,
                    "supply_cap": 5000000000000,
                },
                "0x5678...": {
                    "symbol": "USDC",
                    "symbol_canonical": "usdc",
                    "display_symbol": "USDC",
                    "reserve": {},
                    "decimals": 6,
                    "headroom": 2000000000000,
                    "supply_cap": 10000000000000,
                },
            },
            "notes": [],
        }
        mock_hyperlend_client.get_stable_markets = AsyncMock(return_value=mock_response)

        with patch(
            "wayfinder_paths.adapters.hyperlend_adapter.adapter.HYPERLEND_CLIENT",
            mock_hyperlend_client,
        ):
            success, data = await adapter.get_stable_markets(
                required_underlying_tokens=1000.0,
                buffer_bps=100,
                min_buffer_tokens=100.0,
            )

            assert success
            assert data == mock_response
            mock_hyperlend_client.get_stable_markets.assert_called_once_with(
                required_underlying_tokens=1000.0,
                buffer_bps=100,
                min_buffer_tokens=100.0,
            )

    @pytest.mark.asyncio
    async def test_get_stable_markets_minimal_params(
        self, adapter, mock_hyperlend_client
    ):
        mock_response = {
            "markets": {
                "0x1234...": {
                    "symbol": "USDT",
                    "symbol_canonical": "usdt",
                    "display_symbol": "USDT",
                    "reserve": {},
                    "decimals": 6,
                    "headroom": 1000000000000,
                    "supply_cap": 5000000000000,
                }
            },
            "notes": [],
        }
        mock_hyperlend_client.get_stable_markets = AsyncMock(return_value=mock_response)

        with patch(
            "wayfinder_paths.adapters.hyperlend_adapter.adapter.HYPERLEND_CLIENT",
            mock_hyperlend_client,
        ):
            success, data = await adapter.get_stable_markets()

            assert success
            assert data == mock_response
            mock_hyperlend_client.get_stable_markets.assert_called_once_with(
                required_underlying_tokens=None,
                buffer_bps=None,
                min_buffer_tokens=None,
            )

    @pytest.mark.asyncio
    async def test_get_stable_markets_partial_params(
        self, adapter, mock_hyperlend_client
    ):
        mock_response = {"markets": {}, "notes": []}
        mock_hyperlend_client.get_stable_markets = AsyncMock(return_value=mock_response)

        with patch(
            "wayfinder_paths.adapters.hyperlend_adapter.adapter.HYPERLEND_CLIENT",
            mock_hyperlend_client,
        ):
            success, data = await adapter.get_stable_markets(
                required_underlying_tokens=500.0
            )

            assert success
            assert data == mock_response
            mock_hyperlend_client.get_stable_markets.assert_called_once_with(
                required_underlying_tokens=500.0,
                buffer_bps=None,
                min_buffer_tokens=None,
            )

    @pytest.mark.asyncio
    async def test_get_stable_markets_failure(self, adapter, mock_hyperlend_client):
        mock_hyperlend_client.get_stable_markets = AsyncMock(
            side_effect=Exception("API Error: Connection timeout")
        )

        with patch(
            "wayfinder_paths.adapters.hyperlend_adapter.adapter.HYPERLEND_CLIENT",
            mock_hyperlend_client,
        ):
            success, data = await adapter.get_stable_markets()

            assert success is False
            assert "API Error: Connection timeout" in data

    @pytest.mark.asyncio
    async def test_get_stable_markets_http_error(self, adapter, mock_hyperlend_client):
        mock_hyperlend_client.get_stable_markets = AsyncMock(
            side_effect=Exception("HTTP 404 Not Found")
        )

        with patch(
            "wayfinder_paths.adapters.hyperlend_adapter.adapter.HYPERLEND_CLIENT",
            mock_hyperlend_client,
        ):
            success, data = await adapter.get_stable_markets()

            assert success is False
            assert "404" in data or "Not Found" in data

    @pytest.mark.asyncio
    async def test_get_stable_markets_empty_response(
        self, adapter, mock_hyperlend_client
    ):
        mock_response = {"markets": {}, "notes": []}
        mock_hyperlend_client.get_stable_markets = AsyncMock(return_value=mock_response)

        with patch(
            "wayfinder_paths.adapters.hyperlend_adapter.adapter.HYPERLEND_CLIENT",
            mock_hyperlend_client,
        ):
            success, data = await adapter.get_stable_markets()

            assert success
            assert data == mock_response
            assert len(data.get("markets", {})) == 0

    def test_adapter_type(self, adapter):
        assert adapter.adapter_type == "HYPERLEND"

    def test_strategy_address_optional(self):
        adapter = HyperlendAdapter(config={})
        assert adapter.strategy_wallet_address is None

    @pytest.mark.asyncio
    async def test_get_all_markets_success(self, adapter):
        reserve_keys = UI_POOL_RESERVE_KEYS

        def build_reserve(**overrides):
            base = dict.fromkeys(reserve_keys, 0)
            base.update(
                {
                    "underlyingAsset": "0x0000000000000000000000000000000000000001",
                    "name": "",
                    "symbol": "",
                    "decimals": 18,
                    "usageAsCollateralEnabled": True,
                    "borrowingEnabled": True,
                    "isActive": True,
                    "isFrozen": False,
                    "isPaused": False,
                    "isSiloedBorrowing": False,
                    "aTokenAddress": "0x0000000000000000000000000000000000000002",
                    "variableDebtTokenAddress": "0x0000000000000000000000000000000000000003",
                    "interestRateStrategyAddress": "0x0000000000000000000000000000000000000004",
                    "priceOracle": "0x0000000000000000000000000000000000000005",
                    "flashLoanEnabled": True,
                    "borrowableInIsolation": False,
                    "virtualAccActive": False,
                }
            )
            base.update(overrides)
            return tuple(base[k] for k in reserve_keys)

        reserves = [
            build_reserve(
                underlyingAsset="0x0000000000000000000000000000000000000011",
                symbol="USDC",
                decimals=6,
                baseLTVasCollateral=8000,
                reserveLiquidationThreshold=8500,
                reserveLiquidationBonus=10500,
                reserveFactor=1000,
                liquidityRate=int(0.05 * 10**27),
                variableBorrowRate=int(0.10 * 10**27),
                priceInMarketReferenceCurrency=100000000,
                availableLiquidity=5000000,
                totalScaledVariableDebt=0,
                variableBorrowIndex=10**27,
                supplyCap=0,
            ),
            build_reserve(
                underlyingAsset="0x0000000000000000000000000000000000000022",
                symbol="uSOL",
                decimals=6,
                baseLTVasCollateral=6000,
                reserveLiquidationThreshold=7500,
                reserveLiquidationBonus=11000,
                reserveFactor=2000,
                liquidityRate=int(0.02 * 10**27),
                variableBorrowRate=int(0.08 * 10**27),
                priceInMarketReferenceCurrency=2000000000,
                availableLiquidity=10000000,
                totalScaledVariableDebt=5000000,
                variableBorrowIndex=10**27,
                borrowCap=500,
                supplyCap=100,
                debtCeiling=12345,
                debtCeilingDecimals=2,
            ),
        ]
        base_currency = (100000000, 100000000, 0, 8)  # ref_unit=1e8, ref_usd=1.0

        mock_ui_pool = MagicMock()
        mock_ui_pool.functions.getReservesData = MagicMock(
            return_value=MagicMock(
                call=AsyncMock(return_value=(reserves, base_currency))
            )
        )

        mock_web3 = MagicMock()
        mock_web3.eth.contract = MagicMock(return_value=mock_ui_pool)

        @asynccontextmanager
        async def mock_web3_ctx(_chain_id):
            yield mock_web3

        with patch(
            "wayfinder_paths.adapters.hyperlend_adapter.adapter.web3_from_chain_id",
            mock_web3_ctx,
        ):
            ok, markets = await adapter.get_all_markets()

        assert ok is True
        assert isinstance(markets, list)
        assert len(markets) == 2

        usol = next(m for m in markets if m["symbol"].lower() == "usol")
        assert usol["is_stablecoin"] is False
        assert usol["price_usd"] == 20.0
        assert usol["supply_cap"] == 100
        assert usol["supply_cap_headroom"] == 85000000
        assert usol["ltv_bps"] == 6000
        assert usol["liquidation_threshold_bps"] == 7500
        assert usol["liquidation_bonus_bps"] == 11000
        assert usol["reserve_factor_bps"] == 2000
        assert usol["borrow_cap"] == 500
        assert usol["debt_ceiling"] == 12345
        assert usol["debt_ceiling_decimals"] == 2

    @pytest.mark.asyncio
    async def test_get_stable_markets_with_is_stable_symbol(
        self, adapter, mock_hyperlend_client
    ):
        mock_response = {
            "markets": {
                "0x1234...": {
                    "symbol": "USDT",
                    "symbol_canonical": "usdt",
                    "display_symbol": "USDT",
                    "reserve": {},
                    "decimals": 6,
                    "headroom": 1000000000000,
                    "supply_cap": 5000000000000,
                }
            },
            "notes": [],
        }
        mock_hyperlend_client.get_stable_markets = AsyncMock(return_value=mock_response)

        with patch(
            "wayfinder_paths.adapters.hyperlend_adapter.adapter.HYPERLEND_CLIENT",
            mock_hyperlend_client,
        ):
            success, data = await adapter.get_stable_markets()

            assert success
            assert data == mock_response
            mock_hyperlend_client.get_stable_markets.assert_called_once_with(
                required_underlying_tokens=None,
                buffer_bps=None,
                min_buffer_tokens=None,
            )

    @pytest.mark.asyncio
    async def test_get_assets_view_success(self, adapter, mock_hyperlend_client):
        mock_response = {
            "block_number": 12345,
            "user": "0x0c737cB5934afCb5B01965141F865F795B324080",
            "native_balance_wei": 0,
            "native_balance": 0.0,
            "assets": [
                {
                    "underlying": "0x1234...",
                    "symbol": "USDT",
                    "symbol_canonical": "usdt",
                    "symbol_display": "USDT",
                    "decimals": 6,
                    "a_token": "0x...",
                    "variable_debt_token": "0x...",
                    "usage_as_collateral_enabled": True,
                    "borrowing_enabled": True,
                    "is_active": True,
                    "is_frozen": False,
                    "is_paused": False,
                    "is_siloed_borrowing": False,
                    "is_stablecoin": True,
                    "underlying_wallet_balance": 1000.0,
                    "underlying_wallet_balance_wei": 1000000000,
                    "price_usd": 1.0,
                    "supply": 500.0,
                    "variable_borrow": 0.0,
                    "supply_usd": 500.0,
                    "variable_borrow_usd": 0.0,
                    "supply_apr": 0.05,
                    "supply_apy": 0.05,
                    "variable_borrow_apr": 0.07,
                    "variable_borrow_apy": 0.07,
                }
            ],
            "account_data": {
                "total_collateral_base": 500,
                "total_debt_base": 0,
                "available_borrows_base": 400,
                "current_liquidation_threshold": 8000,
                "ltv": 7500,
                "health_factor_wad": 115792089237316195423570985008687907853269984665640564039457584007913129639935,
                "health_factor": 1.157920892373162e59,
            },
            "base_currency_info": {
                "marketReferenceCurrencyUnit": 100000000,
                "marketReferenceCurrencyPriceInUsd": 100000000,
                "networkBaseTokenPriceInUsd": 0,
                "networkBaseTokenPriceDecimals": 8,
            },
        }
        mock_hyperlend_client.get_assets_view = AsyncMock(return_value=mock_response)

        with patch(
            "wayfinder_paths.adapters.hyperlend_adapter.adapter.HYPERLEND_CLIENT",
            mock_hyperlend_client,
        ):
            success, data = await adapter.get_assets_view(
                user_address="0x0c737cB5934afCb5B01965141F865F795B324080",
            )

            assert success
            assert data == mock_response
            mock_hyperlend_client.get_assets_view.assert_called_once_with(
                user_address="0x0c737cB5934afCb5B01965141F865F795B324080",
            )

    @pytest.mark.asyncio
    async def test_get_assets_view_failure(self, adapter, mock_hyperlend_client):
        mock_hyperlend_client.get_assets_view = AsyncMock(
            side_effect=Exception("API Error: Invalid address")
        )

        with patch(
            "wayfinder_paths.adapters.hyperlend_adapter.adapter.HYPERLEND_CLIENT",
            mock_hyperlend_client,
        ):
            success, data = await adapter.get_assets_view(
                user_address="0x0c737cB5934afCb5B01965141F865F795B324080",
            )

            assert success is False
            assert "API Error: Invalid address" in data

    @pytest.mark.asyncio
    async def test_get_full_user_state_filters_zero_positions(
        self, adapter, mock_hyperlend_client
    ):
        mock_response = {
            "block_number": 12345,
            "user": "0xabc",
            "native_balance_wei": 0,
            "native_balance": 0.0,
            "assets": [
                {
                    "underlying": "0x1",
                    "symbol": "USDC",
                    "symbol_canonical": "usdc",
                    "symbol_display": "USDC",
                    "decimals": 6,
                    "a_token": "0xa",
                    "variable_debt_token": "0xd",
                    "usage_as_collateral_enabled": True,
                    "borrowing_enabled": True,
                    "is_active": True,
                    "is_frozen": False,
                    "is_paused": False,
                    "is_siloed_borrowing": False,
                    "is_stablecoin": True,
                    "underlying_wallet_balance": 0.0,
                    "underlying_wallet_balance_wei": 0,
                    "price_usd": 1.0,
                    "supply": 0.0,
                    "variable_borrow": 0.0,
                    "supply_usd": 0.0,
                    "variable_borrow_usd": 0.0,
                    "supply_apr": 0.0,
                    "supply_apy": 0.0,
                    "variable_borrow_apr": 0.0,
                    "variable_borrow_apy": 0.0,
                },
                {
                    "underlying": "0x2",
                    "symbol": "USDT",
                    "symbol_canonical": "usdt",
                    "symbol_display": "USDT",
                    "decimals": 6,
                    "a_token": "0xa2",
                    "variable_debt_token": "0xd2",
                    "usage_as_collateral_enabled": True,
                    "borrowing_enabled": True,
                    "is_active": True,
                    "is_frozen": False,
                    "is_paused": False,
                    "is_siloed_borrowing": False,
                    "is_stablecoin": True,
                    "underlying_wallet_balance": 0.0,
                    "underlying_wallet_balance_wei": 0,
                    "price_usd": 1.0,
                    "supply": 123.0,
                    "variable_borrow": 0.0,
                    "supply_usd": 123.0,
                    "variable_borrow_usd": 0.0,
                    "supply_apr": 0.05,
                    "supply_apy": 0.05,
                    "variable_borrow_apr": 0.07,
                    "variable_borrow_apy": 0.07,
                },
            ],
            "account_data": {"health_factor": 2.0},
            "base_currency_info": {},
        }
        mock_hyperlend_client.get_assets_view = AsyncMock(return_value=mock_response)

        with patch(
            "wayfinder_paths.adapters.hyperlend_adapter.adapter.HYPERLEND_CLIENT",
            mock_hyperlend_client,
        ):
            ok, state = await adapter.get_full_user_state(account="0xabc")
            assert ok is True
            assert state["protocol"] == "hyperlend"
            assert state["account"] == "0xabc"
            assert len(state["positions"]) == 1
            assert state["positions"][0]["symbol"] == "USDT"

    @pytest.mark.asyncio
    async def test_get_assets_view_empty_response(self, adapter, mock_hyperlend_client):
        mock_response = {
            "block_number": 12345,
            "user": "0x0c737cB5934afCb5B01965141F865F795B324080",
            "native_balance_wei": 0,
            "native_balance": 0.0,
            "assets": [],
            "account_data": {
                "total_collateral_base": 0,
                "total_debt_base": 0,
                "available_borrows_base": 0,
                "current_liquidation_threshold": 0,
                "ltv": 0,
                "health_factor_wad": 0,
                "health_factor": 0.0,
            },
            "base_currency_info": {
                "marketReferenceCurrencyUnit": 100000000,
                "marketReferenceCurrencyPriceInUsd": 100000000,
                "networkBaseTokenPriceInUsd": 0,
                "networkBaseTokenPriceDecimals": 8,
            },
        }
        mock_hyperlend_client.get_assets_view = AsyncMock(return_value=mock_response)

        with patch(
            "wayfinder_paths.adapters.hyperlend_adapter.adapter.HYPERLEND_CLIENT",
            mock_hyperlend_client,
        ):
            success, data = await adapter.get_assets_view(
                user_address="0x0c737cB5934afCb5B01965141F865F795B324080",
            )

            assert success
            assert data == mock_response
            assert len(data.get("assets", [])) == 0
            # New API uses account_data; total_value may not be present
            assert data.get("account_data", {}).get("total_collateral_base") == 0

    @pytest.mark.asyncio
    async def test_lend_native_uses_whype(self, adapter):
        with (
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.encode_call",
                new_callable=AsyncMock,
            ) as mock_encode,
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.send_transaction",
                new_callable=AsyncMock,
            ) as mock_send,
            patch.object(
                adapter, "_record_pool_op", new_callable=AsyncMock
            ) as mock_record,
        ):
            mock_encode.return_value = {"tx": "data"}
            mock_send.return_value = "0xabc"
            mock_record.return_value = None
            adapter.strategy_wallet_signing_callback = AsyncMock()

            ok, txn = await adapter.lend(
                underlying_token="0x0000000000000000000000000000000000000000",
                qty=123,
                chain_id=999,
                native=True,
            )

        assert ok is True
        assert txn == "0xabc"
        mock_encode.assert_awaited_once()
        _, kwargs = mock_encode.await_args
        assert kwargs["target"] == HYPERLEND_WRAPPED_TOKEN_GATEWAY
        assert kwargs["fn_name"] == "depositETH"
        assert kwargs["args"][0] == HYPEREVM_WHYPE
        assert kwargs["value"] == 123

    @pytest.mark.asyncio
    async def test_unlend_native_uses_whype(self, adapter):
        with (
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.encode_call",
                new_callable=AsyncMock,
            ) as mock_encode,
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.send_transaction",
                new_callable=AsyncMock,
            ) as mock_send,
            patch.object(
                adapter, "_record_pool_op", new_callable=AsyncMock
            ) as mock_record,
        ):
            mock_encode.return_value = {"tx": "data"}
            mock_send.return_value = "0xabc"
            mock_record.return_value = None
            adapter.strategy_wallet_signing_callback = AsyncMock()

            ok, txn = await adapter.unlend(
                underlying_token="0x0000000000000000000000000000000000000000",
                qty=123,
                chain_id=999,
                native=True,
            )

        assert ok is True
        assert txn == "0xabc"
        mock_encode.assert_awaited_once()
        _, kwargs = mock_encode.await_args
        assert kwargs["target"] == HYPERLEND_WRAPPED_TOKEN_GATEWAY
        assert kwargs["fn_name"] == "withdrawETH"
        assert kwargs["args"][0] == HYPEREVM_WHYPE

    @pytest.mark.asyncio
    async def test_borrow_erc20_calls_pool_borrow(self, adapter):
        token = "0x0000000000000000000000000000000000000011"
        with (
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.encode_call",
                new_callable=AsyncMock,
            ) as mock_encode,
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.send_transaction",
                new_callable=AsyncMock,
            ) as mock_send,
        ):
            mock_encode.return_value = {"tx": "data"}
            mock_send.return_value = "0xabc"
            adapter.strategy_wallet_signing_callback = AsyncMock()

            ok, txn = await adapter.borrow(
                underlying_token=token,
                qty=123,
                chain_id=999,
                native=False,
            )

        assert ok is True
        assert txn == "0xabc"
        _, kwargs = mock_encode.await_args
        assert kwargs["target"] == HYPERLEND_POOL
        assert kwargs["fn_name"] == "borrow"
        assert kwargs["args"][0].lower() == token.lower()
        assert kwargs["args"][1] == 123
        assert kwargs["args"][2] == 2  # variable rate mode
        assert kwargs["args"][3] == 0  # referral code

    @pytest.mark.asyncio
    async def test_borrow_native_borrows_and_unwraps(self, adapter):
        with (
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.encode_call",
                new_callable=AsyncMock,
            ) as mock_encode,
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.send_transaction",
                new_callable=AsyncMock,
            ) as mock_send,
        ):
            mock_encode.side_effect = [{"tx": "borrow"}, {"tx": "unwrap"}]
            mock_send.side_effect = ["0xborrow", "0xunwrap"]
            adapter.strategy_wallet_signing_callback = AsyncMock()

            ok, res = await adapter.borrow(
                underlying_token="0x0000000000000000000000000000000000000000",
                qty=123,
                chain_id=999,
                native=True,
            )

        assert ok is True
        assert res == {"borrow_tx": "0xborrow", "unwrap_tx": "0xunwrap"}

        assert mock_encode.await_count == 2
        first_kwargs = mock_encode.await_args_list[0].kwargs
        assert first_kwargs["target"] == HYPERLEND_POOL
        assert first_kwargs["fn_name"] == "borrow"
        assert first_kwargs["args"][0] == HYPEREVM_WHYPE
        assert first_kwargs["args"][1] == 123
        assert first_kwargs["args"][2] == 2  # variable rate mode
        assert first_kwargs["args"][3] == 0  # referral code
        assert first_kwargs["args"][4] == adapter.strategy_wallet_address

        second_kwargs = mock_encode.await_args_list[1].kwargs
        assert second_kwargs["target"] == HYPEREVM_WHYPE
        assert second_kwargs["fn_name"] == "withdraw"
        assert second_kwargs["args"][0] == 123

    @pytest.mark.asyncio
    async def test_repay_erc20_approves_and_calls_pool_repay(self, adapter):
        token = "0x0000000000000000000000000000000000000011"
        with (
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.ensure_allowance",
                new_callable=AsyncMock,
            ) as mock_allow,
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.encode_call",
                new_callable=AsyncMock,
            ) as mock_encode,
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.send_transaction",
                new_callable=AsyncMock,
            ) as mock_send,
        ):
            mock_allow.return_value = (True, "ok")
            mock_encode.return_value = {"tx": "data"}
            mock_send.return_value = "0xabc"
            adapter.strategy_wallet_signing_callback = AsyncMock()

            ok, txn = await adapter.repay(
                underlying_token=token,
                qty=123,
                chain_id=999,
                native=False,
                repay_full=True,
            )

        assert ok is True
        assert txn == "0xabc"
        mock_allow.assert_awaited_once()
        _, allow_kwargs = mock_allow.await_args
        assert allow_kwargs["spender"] == HYPERLEND_POOL
        assert allow_kwargs["approval_amount"] == MAX_UINT256

        _, kwargs = mock_encode.await_args
        assert kwargs["target"] == HYPERLEND_POOL
        assert kwargs["fn_name"] == "repay"
        assert kwargs["args"][0].lower() == token.lower()
        assert kwargs["args"][1] == MAX_UINT256
        assert kwargs["args"][2] == 2  # variable rate mode

    @pytest.mark.asyncio
    async def test_repay_native_calls_gateway_repayeth_with_value(self, adapter):
        with (
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.encode_call",
                new_callable=AsyncMock,
            ) as mock_encode,
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.send_transaction",
                new_callable=AsyncMock,
            ) as mock_send,
        ):
            mock_encode.return_value = {"tx": "data"}
            mock_send.return_value = "0xabc"
            adapter.strategy_wallet_signing_callback = AsyncMock()

            ok, txn = await adapter.repay(
                underlying_token="0x0000000000000000000000000000000000000000",
                qty=123,
                chain_id=999,
                native=True,
                repay_full=False,
            )

        assert ok is True
        assert txn == "0xabc"
        _, kwargs = mock_encode.await_args
        assert kwargs["target"] == HYPERLEND_WRAPPED_TOKEN_GATEWAY
        assert kwargs["fn_name"] == "repayETH"
        assert kwargs["args"][0] == HYPEREVM_WHYPE
        assert kwargs["args"][1] == 123
        assert kwargs["args"][2] == adapter.strategy_wallet_address
        assert kwargs["value"] == 123

    @pytest.mark.asyncio
    async def test_repay_native_full_repays_max_uint_with_buffer(self, adapter):
        mock_ui_pool = MagicMock()
        reserves = [
            {
                "underlyingAsset": HYPEREVM_WHYPE,
                "variableDebtTokenAddress": "0x00000000000000000000000000000000000000dE",
            }
        ]
        base_currency = (0, 0, 0, 0)
        mock_ui_pool.functions.getReservesData = MagicMock(
            return_value=MagicMock(
                call=AsyncMock(return_value=(reserves, base_currency))
            )
        )

        mock_web3 = MagicMock()
        mock_web3.eth.contract = MagicMock(return_value=mock_ui_pool)

        @asynccontextmanager
        async def mock_web3_ctx(_chain_id):
            yield mock_web3

        with (
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.web3_from_chain_id",
                mock_web3_ctx,
            ),
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.get_token_balance",
                new_callable=AsyncMock,
            ) as mock_bal,
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.encode_call",
                new_callable=AsyncMock,
            ) as mock_encode,
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.send_transaction",
                new_callable=AsyncMock,
            ) as mock_send,
        ):
            mock_bal.side_effect = [1000, 10000]  # debt, native balance
            mock_encode.return_value = {"tx": "data"}
            mock_send.return_value = "0xabc"
            adapter.strategy_wallet_signing_callback = AsyncMock()

            ok, txn = await adapter.repay(
                underlying_token="0x0000000000000000000000000000000000000000",
                qty=0,
                chain_id=999,
                native=True,
                repay_full=True,
            )

        assert ok is True
        assert txn == "0xabc"
        assert mock_bal.await_count == 2
        _, kwargs = mock_encode.await_args
        assert kwargs["target"] == HYPERLEND_WRAPPED_TOKEN_GATEWAY
        assert kwargs["fn_name"] == "repayETH"
        assert kwargs["args"][0] == HYPEREVM_WHYPE
        assert kwargs["args"][1] == MAX_UINT256
        assert kwargs["args"][2] == adapter.strategy_wallet_address
        assert kwargs["value"] == 1001  # 1000 + 1 wei buffer

    @pytest.mark.asyncio
    async def test_repay_native_full_no_debt_is_noop(self, adapter):
        mock_ui_pool = MagicMock()
        reserves = [
            {
                "underlyingAsset": HYPEREVM_WHYPE,
                "variableDebtTokenAddress": "0x00000000000000000000000000000000000000dE",
            }
        ]
        base_currency = (0, 0, 0, 0)
        mock_ui_pool.functions.getReservesData = MagicMock(
            return_value=MagicMock(
                call=AsyncMock(return_value=(reserves, base_currency))
            )
        )

        mock_web3 = MagicMock()
        mock_web3.eth.contract = MagicMock(return_value=mock_ui_pool)

        @asynccontextmanager
        async def mock_web3_ctx(_chain_id):
            yield mock_web3

        with (
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.web3_from_chain_id",
                mock_web3_ctx,
            ),
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.get_token_balance",
                new_callable=AsyncMock,
            ) as mock_bal,
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.encode_call",
                new_callable=AsyncMock,
            ) as mock_encode,
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.send_transaction",
                new_callable=AsyncMock,
            ) as mock_send,
        ):
            mock_bal.return_value = 0

            ok, txn = await adapter.repay(
                underlying_token="0x0000000000000000000000000000000000000000",
                qty=0,
                chain_id=999,
                native=True,
                repay_full=True,
            )

        assert ok is True
        assert txn is None
        mock_encode.assert_not_awaited()
        mock_send.assert_not_awaited()

    @pytest.mark.asyncio
    async def test_repay_native_full_requires_sufficient_balance(self, adapter):
        mock_ui_pool = MagicMock()
        reserves = [
            {
                "underlyingAsset": HYPEREVM_WHYPE,
                "variableDebtTokenAddress": "0x00000000000000000000000000000000000000dE",
            }
        ]
        base_currency = (0, 0, 0, 0)
        mock_ui_pool.functions.getReservesData = MagicMock(
            return_value=MagicMock(
                call=AsyncMock(return_value=(reserves, base_currency))
            )
        )

        mock_web3 = MagicMock()
        mock_web3.eth.contract = MagicMock(return_value=mock_ui_pool)

        @asynccontextmanager
        async def mock_web3_ctx(_chain_id):
            yield mock_web3

        with (
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.web3_from_chain_id",
                mock_web3_ctx,
            ),
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.get_token_balance",
                new_callable=AsyncMock,
            ) as mock_bal,
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.encode_call",
                new_callable=AsyncMock,
            ) as mock_encode,
            patch(
                "wayfinder_paths.adapters.hyperlend_adapter.adapter.send_transaction",
                new_callable=AsyncMock,
            ) as mock_send,
        ):
            mock_bal.side_effect = [1000, 999]  # debt, native balance

            ok, err = await adapter.repay(
                underlying_token="0x0000000000000000000000000000000000000000",
                qty=0,
                chain_id=999,
                native=True,
                repay_full=True,
            )

        assert ok is False
        assert "insufficient HYPE balance for repay_full" in str(err)
        mock_encode.assert_not_awaited()
        mock_send.assert_not_awaited()



from __future__ import annotations

from abc import ABC
from typing import Any

from loguru import logger


class BaseAdapter(ABC):
    adapter_type: str | None = None

    def __init__(self, name: str, config: dict[str, Any] | None = None):
        self.name = name
        self.config = config or {}
        self.logger = logger.bind(adapter=self.__class__.__name__)

    async def close(self) -> None:
        pass
